/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * bitmap_test.c
 * -------------
 *
 * A test of bitmaps
 */

#include "impulse/rows/bitmap.h"
#include "impulse/util/memory.h"
#include "impulse/util/event.h"

#include "bitmap_test.h"

/*                            ======--------======                           */

static void bitmap_test_setup(ROWS_widget *widget);
static int  bitmap_test_ID   = -1;
static int  bitmap_test_type = -1;

static void bitmap_test_destroy(void **private);

static impulse_bool  bitmap_test_redraw(ROWS_child            widget,
					ROWS_child_event_num  event_number,
					ROWS_child_event     *event,
					void                     *data);
static impulse_bool  bitmap_test_crossing (ROWS_child            widget,
					   ROWS_child_event_num  event_number,
					   ROWS_child_event     *event,
					   void                     *data);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget bitmap_test_create(void)
{
  ROWS_widget *wid;

  if (bitmap_test_ID   == -1)
    bitmap_test_ID   = ROWS_child_next_identifier();

  wid = bitmap_test_initialise(bitmap_test_ID);

  return *wid;
}

ROWS_widget *bitmap_test_initialise(int id)
{
  ROWS_widget *widget;

  if (bitmap_test_ID   == -1)
    bitmap_test_ID   = ROWS_child_next_identifier();
  if (bitmap_test_type == -1)
    bitmap_test_type = ROWS_widget_next_type();

  widget = ROWS_child_initialise(id, NULL);
  ROWS_widget_set_type(widget, bitmap_test_ID, bitmap_test_type);
  ROWS_widget_set_destructor(widget, bitmap_test_ID, bitmap_test_destroy);

  bitmap_test_setup(widget);

  return widget;
}

int bitmap_test_next_identifier(void)
{
  if (bitmap_test_ID==-1)
    bitmap_test_ID = ROWS_child_next_identifier();

  return bitmap_test_ID + 1;
}

bitmap_test bitmap_test_cast(void **widget)
{
  if (bitmap_test_ID   == -1)
    bitmap_test_ID   = ROWS_child_next_identifier();
  if (bitmap_test_type == -1)
    bitmap_test_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, bitmap_test_ID, bitmap_test_type);
}

struct bitmap_test {
  ROWS_bitmap bitmap;

  int timer;

  int r1,  g1,  b1;
  int r1d, g1d, b1d;

  int r2,  g2,  b2;
  int r2d, g2d, b2d;
};

void bitmap_test_makeimage(ROWS_bitmap bitmap,
			   int r1, int g1, int b1,
			   int r2, int g2, int b2)
{
  impulse_int   *data;
  ROWS_intpoint  size;
  float          r,     g,     b;
  float          rstep, gstep, bstep;
  int            x,y,pos;
  
  size = ROWS_bitmap_get_size(bitmap);
  data = ROWS_bitmap_get_data(bitmap);

  rstep = ((float)(r2-r1))/(float)size.x;
  gstep = ((float)(g2-g1))/(float)size.x;
  bstep = ((float)(b2-b1))/(float)size.x;
  
  pos = 0;
  for (y=0; y<size.y; y++)
    {
      r = r1;
      g = g1+((float)y*gstep);
      b = b1;
      for (x=0; x<size.x; x++)
	{
	  r += rstep;
	  b += bstep;

	  data[pos] = ((int)r)|(((int)g)<<8)|(((int)b)<<16);

	  pos++;
	}

      pos++;
    }

  /* Mark the square as modified */
  ROWS_bitmap_modified(bitmap);
}

#define DCHECK(x, y) if (x>255) { x = 255; y = -y; } \
if (x<0) { x = 0; y = -y; }

impulse_bool bitmap_ticker(impulse_event *event,
			   void *data)
{
  struct bitmap_test **private;
  ROWS_drawable        drawable;
  ROWS_rect            box;
  ROWS_widget          child;
  
  private = (struct bitmap_test **) data;

  child = ROWS_child_cast((void **) private);

  drawable = ROWS_child_get_drawable(child);
  box.size = *ROWS_child_get_size(ROWS_box_cast(child));
  box.pos  = *ROWS_child_get_offset(ROWS_child_cast(child));

  ROWS_child_lock(child);
  
  /* Create a new image */
  bitmap_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			(*private)->r1, (*private)->g1, (*private)->b1,
			(*private)->r2, (*private)->g2, (*private)->b2);

  ROWS_child_unlock(child);

  ROWS_widget_lock(drawable);

  /* Draw the bitmap... */
  ROWS_drawable_clip_push(drawable);
  ROWS_drawable_clip_clear(drawable);
  ROWS_drawable_clip_union(drawable, &box);
  ROWS_drawable_clip_set(drawable);
  ROWS_drawable_draw_bitmap(drawable,
			    ROWS_bitmap_cast((*private)->bitmap),
			    box.pos.x+(box.size.x/2)-PIXEL_TO_MICRONS(20),
			    box.pos.y+(box.size.y/2)-PIXEL_TO_MICRONS(20));
  ROWS_drawable_clip_pop(drawable);
  ROWS_drawable_sync(drawable);
  
  ROWS_widget_unlock(drawable);

  /* Update the values to make it animate... */
  (*private)->r1 += (*private)->r1d;
  (*private)->g1 += (*private)->g1d;
  (*private)->b1 += (*private)->b1d;
  (*private)->r2 += (*private)->r2d;
  (*private)->g2 += (*private)->g2d;
  (*private)->b2 += (*private)->b2d;

  DCHECK((*private)->r1, (*private)->r1d)
  DCHECK((*private)->g1, (*private)->g1d)
  DCHECK((*private)->b1, (*private)->b1d)
  DCHECK((*private)->r2, (*private)->r2d)
  DCHECK((*private)->g2, (*private)->g2d)
  DCHECK((*private)->b2, (*private)->b2d)
}

void bitmap_test_setup(ROWS_widget *widget)
{
  struct bitmap_test **private;
  impulse_int             *data;
  impulse_unistring tfile;
  int x,y;

  private  = (struct bitmap_test **) ROWS_widget_private(widget, bitmap_test_ID);
  *private = impulse_alloc(sizeof(struct bitmap_test));

  (*private)->bitmap = ROWS_bitmap_create(PIXEL_TO_MICRONS(40),
					  PIXEL_TO_MICRONS(40),
					  32,
					  0);

  tfile = impulse_unistring_create_utf(PREFIX_DIR, 8);
  impulse_unistring_append_utf(tfile, "/demos/rows/test.png",8);
  ROWS_bitmap_load(ROWS_bitmap_cast((*private)->bitmap), tfile);
  impulse_unistring_destroy(tfile);
  
  data = ROWS_bitmap_get_data(ROWS_bitmap_cast((*private)->bitmap));

  /* Draw a colour square (red in top left, white in bottom right) */

  bitmap_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			255,0,0,
			255,255,255);

  (*private)->r1 = 255;
  (*private)->g1 = 0;
  (*private)->b1 = 0;
  (*private)->r2 = 255;
  (*private)->g2 = 255;
  (*private)->b2 = 255;

  (*private)->r1d = -3;
  (*private)->g1d = 1;
  (*private)->b1d = 2;
  (*private)->r2d = -1;
  (*private)->g2d = -2;
  (*private)->b2d = -3;
  
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_EXPOSURE,
			    bitmap_test_redraw,
			    NULL);
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_CROSSING,
			    bitmap_test_crossing,
			    NULL);

  (*private)->timer = impulse_event_timer();
  impulse_event_register_ticker((*private)->timer,
				0, 100000,
				bitmap_ticker,
				private);
}

void bitmap_test_destroy(void **data)
{
  struct bitmap_test **private;

  private = (struct bitmap_test **) data;

  impulse_event_remove_timer((*private)->timer);
  ROWS_widget_destroy((*private)->bitmap);
  /* impulse_free(*private); */
}

impulse_bool bitmap_test_crossing(ROWS_child            widget,
				  ROWS_child_event_num  event_number,
				  ROWS_child_event     *event,
				  void                 *data)
{
  if (event->crossing.entered)
    ROWS_child_set_activation(widget,
			      ROWS_STYLE_HIGHLIGHT);
  else
    ROWS_child_set_activation(widget,
			      ROWS_STYLE_NORMAL);
  return 0;
}
     

impulse_bool bitmap_test_redraw(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data)
{
  ROWS_drawable            drawable;
  ROWS_style              style;
  ROWS_rect                box;
  ROWS_style_active_type   type;
  struct bitmap_test      **private;

  private = (struct bitmap_test **) bitmap_test_cast(widget);
  type = ROWS_child_get_activation(widget);

  drawable = ROWS_child_get_drawable(widget);

  style = ROWS_child_get_style(widget);
  box.size = *ROWS_child_get_size(ROWS_box_cast(widget));
  box.pos  = *ROWS_child_get_offset(ROWS_child_cast(widget));

  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERSHADOW],
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y,
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y+box.size.y-1);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERSHADOW],
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y+box.size.y-PIXEL_MULTIPLIER,
		       box.pos.x,
		       box.pos.y+box.size.y-PIXEL_MULTIPLIER);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERHILITE],
		       box.pos.x, box.pos.y+box.size.y-PIXEL_MULTIPLIER,
		       box.pos.x, box.pos.y);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERHILITE],
		       box.pos.x, box.pos.y,
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER, box.pos.y);

  ROWS_drawable_draw_bitmap(drawable,
			    ROWS_bitmap_cast((*private)->bitmap),
			    box.pos.x+(box.size.x/2)-PIXEL_TO_MICRONS(20),
			    box.pos.y+(box.size.y/2)-PIXEL_TO_MICRONS(20));
  
  return 1;
}
