/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * colour_test.c
 * -------------
 *
 * A test of bitmaps
 */

#include "impulse/rows/bitmap.h"
#include "impulse/util/memory.h"
#include "impulse/util/event.h"

#include "colour_test.h"

/*                            ======--------======                           */

static void colour_test_setup(ROWS_widget *widget,
			      ROWS_colour_type type);
static int  colour_test_ID   = -1;
static int  colour_test_type = -1;

static void colour_test_destroy(void **private);

static impulse_bool  colour_test_redraw(ROWS_child            widget,
					ROWS_child_event_num  event_number,
					ROWS_child_event     *event,
					void                     *data);
static impulse_bool  colour_test_crossing (ROWS_child            widget,
					   ROWS_child_event_num  event_number,
					   ROWS_child_event     *event,
					   void                     *data);

/*                            ======--------======                           */

/* Standard widget functions */

ROWS_widget colour_test_create(ROWS_colour_type col_type)
{
  ROWS_widget *wid;

  if (colour_test_ID   == -1)
    colour_test_ID   = ROWS_child_next_identifier();

  wid = colour_test_initialise(colour_test_ID, col_type);

  return *wid;
}

ROWS_widget *colour_test_initialise(int id, ROWS_colour_type col_type)
{
  ROWS_widget *widget;

  if (colour_test_ID   == -1)
    colour_test_ID   = ROWS_child_next_identifier();
  if (colour_test_type == -1)
    colour_test_type = ROWS_widget_next_type();

  widget = ROWS_child_initialise(id, NULL);
  ROWS_widget_set_type(widget, colour_test_ID, colour_test_type);
  ROWS_widget_set_destructor(widget, colour_test_ID, colour_test_destroy);

  colour_test_setup(widget, col_type);

  return widget;
}

int colour_test_next_identifier(void)
{
  if (colour_test_ID==-1)
    colour_test_ID = ROWS_child_next_identifier();

  return colour_test_ID + 1;
}

colour_test colour_test_cast(void **widget)
{
  if (colour_test_ID   == -1)
    colour_test_ID   = ROWS_child_next_identifier();
  if (colour_test_type == -1)
    colour_test_type = ROWS_widget_next_type();

  return ROWS_widget_cast_misc(widget, colour_test_ID, colour_test_type);
}

struct colour_test {
  ROWS_bitmap    bitmap;
  ROWS_gamut gamut;
};

void colour_test_makeimage(ROWS_bitmap bitmap,
			   int xb, int yb,
			   int xt, int yt,
			   int z,
			   ROWS_gamut gamut,
			   ROWS_colour_type type)
{
  impulse_int   *data;
  ROWS_intpoint  size;
  float          xstep, ystep;
  int            x,y,pos;
  ROWS_colour_desc xyz;
  ROWS_colour_desc rgb;
  
  size = ROWS_bitmap_get_size(bitmap);
  data = ROWS_bitmap_get_data(bitmap);

  xstep = ((float)(xt-xb))/((float) size.x);
  ystep = ((float)(yt-yb))/((float) size.y);
  
  pos = 0;
  for (y=0; y<size.y; y++)
    {
      for (x=0; x<size.x; x++)
	{
	  int r,g,b;

	  switch (type)
	    {
	    case ROWS_colour_XYZ:
	      xyz.type = type;
	      xyz.data.xyz.x = xb+((float)x*xstep);
	      xyz.data.xyz.y = z;
	      xyz.data.xyz.z = yb+((float)y*ystep);
	      break;

	    case ROWS_colour_Yxy:
	      xyz.type = type;
	      xyz.data.Yxy.x = xb+((float)x*xstep);
	      xyz.data.Yxy.y = yb+((float)y*ystep);
	      xyz.data.Yxy.Y = z;
	      break;

	    case ROWS_colour_YUV:
	    case ROWS_colour_YUpVp:
	      xyz.type = type;
	      xyz.data.yuv.y = z;
	      xyz.data.yuv.u = xb+((float)x*xstep);
	      xyz.data.yuv.v = yb+((float)y*ystep);
	      break;
	      
	    case ROWS_colour_LAB:
	    case ROWS_colour_LUV:
	      xyz.type = type;
	      xyz.data.lab.a = xb+((float)x*xstep);
	      xyz.data.lab.b = yb+((float)y*ystep);
	      xyz.data.lab.l = z;
	    }
	      
	  rgb = ROWS_colour_convert(xyz, ROWS_colour_RGB, gamut);
	  r = rgb.data.rgb.r>>4;
	  g = rgb.data.rgb.g>>4;
	  b = rgb.data.rgb.b>>4;

	  /*
	  r = r<0?0:r;
	  r = r>255?255:r;
	  g = g<0?0:g;
	  g = g>255?255:g;
	  b = b<0?0:b;
	  b = b>255?255:b;
	  */
	  
	  if (r<0 || r>255 || g<0 || g>255 || b<0 || b>255)
	    {
	      r = g = b = 0;
	    }

	  data[pos] = ((int)r)|(((int)g)<<8)|(((int)b)<<16);

	  pos++;
	}

      pos++;
    }

  /* Mark the square as modified */
  ROWS_bitmap_modified(bitmap);
}

#define DCHECK(x, y) if (x>255) { x = 255; y = -y; } \
if (x<0) { x = 0; y = -y; }

void colour_test_setup(ROWS_widget *widget,
		       ROWS_colour_type type)
{
  struct colour_test **private;
  impulse_int             *data;
  impulse_unistring tfile;

  private  = (struct colour_test **) ROWS_widget_private(widget, colour_test_ID);
  *private = impulse_alloc(sizeof(struct colour_test));

  (*private)->bitmap = ROWS_bitmap_create(PIXEL_TO_MICRONS(200),
					  PIXEL_TO_MICRONS(200),
					  32,
					  0);
  (*private)->gamut = ROWS_colour_make_gamut(/* Col X,Y coordinates */
					     2499 /* 0.61 */, 1434 /* 0.35 */,
					     1188 /* 0.29 */, 2417 /* 0.59 */,
					     614 /* 0.15 */, 246  /* 0.06 */,
					     /* Chromacity */
					     /*
					      * Assuming white is equal mix of
					      * all colours (not unreasonable)
					      */
					     1365,
					     1365,
					     1365,
					     2.2);

  tfile = impulse_unistring_create_utf(PREFIX_DIR, 8);
  impulse_unistring_append_utf(tfile, "/demos/rows/test.png",8);
  ROWS_bitmap_load(ROWS_bitmap_cast((*private)->bitmap), tfile);
  impulse_unistring_destroy(tfile);
  
  data = ROWS_bitmap_get_data(ROWS_bitmap_cast((*private)->bitmap));

  /* Draw a colour square */

  switch (type)
    {
    case ROWS_colour_LAB:
    case ROWS_colour_LUV:
      colour_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			    -4096,-4096,
			    4096, 4096,
			    3000, (*private)->gamut, type);
      break;

    case ROWS_colour_YUV:
      colour_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			    400, 1000,
			    1700, 1600,
			    512, (*private)->gamut, type);
      break;

    case ROWS_colour_YUpVp:
      colour_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			    400, 1500,
			    1700, 2400,
			    512, (*private)->gamut, type);
      break;
 
    case ROWS_colour_Yxy:
      colour_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			    700, 800,
			    2550, 2450,
			    512, (*private)->gamut, type);
      break;
   
    default:
      colour_test_makeimage(ROWS_bitmap_cast((*private)->bitmap),
			    0,0,
			    1365, 1365,
			    682, (*private)->gamut, type);
      break;
    }
 
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_EXPOSURE,
			    colour_test_redraw,
			    NULL);
  ROWS_child_event_register(ROWS_child_cast(*widget),
			    ROWS_CHILD_CROSSING,
			    colour_test_crossing,
			    NULL);

  {
    ROWS_colour_desc in,out,out2;
    ROWS_gamut   gam;
    int x, y;
    
    gam = (*private)->gamut;
    
    in.type = ROWS_colour_RGB;
    in.data.rgb.r = 4095;
    in.data.rgb.g = 4095;
    in.data.rgb.b = 4095;

    out = ROWS_colour_convert(in, ROWS_colour_XYZ, gam);
    out2 = ROWS_colour_convert(out, ROWS_colour_RGB, gam);

    printf("%i, %i, %i (RGB) -> %i, %i, %i (CIE) -> %i, %i, %i (RGB)\n",
	   in.data.rgb.r,
	   in.data.rgb.g,
	   in.data.rgb.b,
	   out.data.xyz.x,
	   out.data.xyz.y,
	   out.data.xyz.z,
	   out2.data.rgb.r,
	   out2.data.rgb.g,
	   out2.data.rgb.b);
  }
}

void colour_test_destroy(void **data)
{
  struct colour_test **private;

  private = (struct colour_test **) data;

  ROWS_widget_destroy((*private)->bitmap);
  /* impulse_free(*private); */
}

impulse_bool colour_test_crossing(ROWS_child            widget,
				  ROWS_child_event_num  event_number,
				  ROWS_child_event     *event,
				  void                 *data)
{
  if (event->crossing.entered)
    ROWS_child_set_activation(widget,
			      ROWS_STYLE_HIGHLIGHT);
  else
    ROWS_child_set_activation(widget,
			      ROWS_STYLE_NORMAL);
  return 0;
}
     

impulse_bool colour_test_redraw(ROWS_child            widget,
				ROWS_child_event_num  event_number,
				ROWS_child_event     *event,
				void                 *data)
{
  ROWS_drawable            drawable;
  ROWS_style              style;
  ROWS_rect                box;
  ROWS_style_active_type   type;
  struct colour_test      **private;

  private = (struct colour_test **) colour_test_cast(widget);
  type = ROWS_child_get_activation(widget);

  drawable = ROWS_child_get_drawable(widget);

  style = ROWS_child_get_style(widget);
  box.size = *ROWS_child_get_size(ROWS_box_cast(widget));
  box.pos  = *ROWS_child_get_offset(ROWS_child_cast(widget));

  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERSHADOW],
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y,
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y+box.size.y-1);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERSHADOW],
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER,
		       box.pos.y+box.size.y-PIXEL_MULTIPLIER,
		       box.pos.x,
		       box.pos.y+box.size.y-PIXEL_MULTIPLIER);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERHILITE],
		       box.pos.x, box.pos.y+box.size.y-PIXEL_MULTIPLIER,
		       box.pos.x, box.pos.y);
  ROWS_brush_draw_line(style->brush[type][ROWS_STYLE_BORDERHILITE],
		       box.pos.x, box.pos.y,
		       box.pos.x+box.size.x-PIXEL_MULTIPLIER, box.pos.y);

  ROWS_drawable_draw_bitmap(drawable,
			    ROWS_bitmap_cast((*private)->bitmap),
			    box.pos.x+(box.size.x/2)-PIXEL_TO_MICRONS(100),
			    box.pos.y+(box.size.y/2)-PIXEL_TO_MICRONS(100));
  
  return 1;
}
