/*
 **  test_philo.c -- Pth test application showing the 5 philosopher problem
**
**  Copyright (c) 1999 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of GNU Pth, a non-preemptive thread scheduling
**  library which can be found at http://www.gnu.org/software/pth/.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA, or contact Ralf S. Engelschall <rse@engelschall.com>.
*/

/*
 *  E.W. Dijkstra, ``Hierarchical Ordering of Sequential Processes.'',
 *  Acta Informatica 1, 1971, 115-138. 
 */

/*
 * Modified to use as a test program for the impulse threading routines
 * by Daniel Berrange (dan@impulse.org.uk)
 */

#include "impulse/util/thread.h"
#include "impulse/util/event.h"
#include "impulse/util/main.h"
#include "impulse/util/memory.h"
#include "impulse/util/overseer.h"
#if USE_THREADS
#include <stdlib.h>

#include <stdio.h>

#define PHILNUM 5

typedef enum {
    thinking,
    hungry,
    eating
} philstat;

static char *philstatstr[3] = {
    "\033[31mthinking\033[0m",
    "\033[32mhungry\033[0m  ",
    "\033[33mEATING\033[0m  "
};

typedef struct tablestruct {
  impulse_thread       tid[PHILNUM];
  int         self[PHILNUM];
  impulse_mutex mutex;
  impulse_cond condition[PHILNUM];
  philstat    status[PHILNUM];
} table;

static table *tab;


static void printstate(void)
{
    int i;

    for (i = 0; i < PHILNUM; i++)
        printf("| %s ", philstatstr[(int)(tab->status)[i]]);
    printf("|\n");
    return;
}

static int test(int i)
{
    if (   ((tab->status)[i] == hungry)
        && ((tab->status)[(i + 1) % PHILNUM] != eating)
        && ((tab->status)[(i - 1 + PHILNUM) % PHILNUM] != eating)) {
        (tab->status)[i] = eating;
        impulse_cond_signal((tab->condition)[i]);
        return 1;
    }
    return 0;
}

static void pickup(int k)
{
    impulse_mutex_lock(tab->mutex);
    (tab->status)[k] = hungry;
    printstate();
    if (!test(k))
        impulse_cond_wait((tab->condition)[k], tab->mutex);
    printstate();
    impulse_mutex_unlock(tab->mutex);
    return;
}

static void putdown(int k)
{
    impulse_mutex_lock(tab->mutex);
    (tab->status)[k] = thinking;
    printstate();
    test((k + 1) % PHILNUM);
    test((k - 1 + PHILNUM) % PHILNUM);
    impulse_mutex_unlock(tab->mutex);
    return;
}

static void *philosopher(void *_who)
{
  int *who = (int *)_who;
  
  for (;;) {
    impulse_thread_nap((2+(int) (2.0*rand()/(RAND_MAX+1.0))),0);
    pickup((*who));
    impulse_thread_nap((2+(int) (2.0*rand()/(RAND_MAX+1.0))),0);
    putdown((*who));
  }
  return NULL;
}
#endif


impulse_bool quitter(impulse_event *ev,
		     void *data)
{
  impulse_quit();
  return 0;
}


impulse_int main(impulse_int argc, impulse_char *argv[])
{
#if USE_THREADS
  impulse_int i;
  impulse_int timer;
  
  impulse_init(argc, argv, "thread");
  impulse_overseer_oversee();
  impulse_thread_init();
  impulse_command_parse();
  
  /* display test program header */
  printf("This is TEST_PHILO, a Pth test showing the Five Dining Philosophers\n");
  printf("\n");
  printf("This is a demonstration showing the famous concurrency problem of the\n");
  printf("Five Dining Philosophers as analysed 1965 by E.W.Dijkstra:\n");
  printf("\n");
  printf("Five philosophers are sitting around a round table, each with a bowl of\n");
  printf("Chinese food in front of him. Between periods of talking they may start\n");
  printf("eating whenever they want to, with their bowls being filled frequently.\n");
  printf("But there are only five chopsticks available, one each to the left of\n");
  printf("each bowl - and for eating Chinese food one needs two chopsticks. When\n");
  printf("a philosopher wants to start eating, he must pick up the chopstick to\n");
  printf("the left of his bowl and the chopstick to the right of his bowl. He\n");
  printf("may find, however, that either one (or even both) of the chopsticks is\n");
  printf("unavailable as it is being used by another philosopher sitting on his\n");
  printf("right or left, so he has to wait.\n");
  printf("\n");
  printf("This situation shows classical contention under concurrency (the\n");
  printf("philosophers want to grab the chopsticks) and the possibility of a\n");
  printf("deadlock (all philosophers wait that the chopstick to their left becomes\n");
  printf("available).\n");
  printf("\n");
  printf("The demonstration runs max. 60 seconds. To stop before, press CTRL-C.\n");
  printf("\n");
  printf("+----P1----+----P2----+----P3----+----P4----+----P5----+\n");
  
  /* initialize the control table */
  tab = (table *)impulse_alloc(sizeof(table));
  tab->mutex = impulse_mutex_create();
  for (i = 0; i < PHILNUM; i++) {
    (tab->self)[i] = i;
    (tab->status)[i] = thinking;
    (tab->condition)[i] = impulse_cond_create();
  }
  
  /* spawn the philosopher threads */
  for (i = 0; i < PHILNUM; i++) {
    (tab->tid)[i] = impulse_thread_create_detach(philosopher, &((tab->self)[i]));
  }

  timer = impulse_event_timer();
  impulse_event_register_ticker(timer, 60, 0, quitter, NULL);
  impulse_main();
  
  printf("+----------+----------+----------+----------+----------+\n");
  fflush(stdout);
#else
  printf("Ok, who's the spoon trying to run the threads test program\n");
  printf("without first enabling thread support in libImpUtil?\n");
#endif
  impulse_thread_cleanup();
  impulse_cleanup();
  return 0;
}

