/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * message.c
 * ---------
 *
 * Abstract message handling
 */


#include "../../../config.h"

#include "impulse/comms/message.h"
#include "impulse/util/memory.h"

#include <stdarg.h>


impulse_message impulse_message_create(impulse_ubyte service,
				       impulse_ubyte opcode,
				       impulse_ushort length)
{
  impulse_message message = impulse_alloc(sizeof(struct _impulse_message));
  
  message->header.service = service;
  message->header.opcode = opcode;
  message->header.length = length;
  
  message->payload = impulse_alloc(length);
  message->offset = message->payload;
  
  return message;
}


void impulse_message_destroy(impulse_message message)
{
  impulse_free(message->payload);
  impulse_free(message);
}


#define impulse_message_write_item(type) \
{ \
  type v = va_arg(args, type); \
  *((type *)message->offset)++ = type##_be(v); \
}

#define impulse_message_write_list(type) \
{ \
  type *v = va_arg(args, type*); \
  for (i = 0 ; i < len ; i++) \
  {\
    *((type *)message->offset)++ = type##_be(*v); \
    v++;\
  } \
}


void impulse_message_writef(impulse_message message,
			    impulse_uint nformatters,
			    impulse_message_formatter *formatters,
			    impulse_char *format,
			    ...)
{
  va_list args;
  va_start(args, format);

  while(*format)
  {
    switch (*format)
    {
     case IMPULSE_MESSAGE_BYTE:
      impulse_message_write_item(impulse_byte)
      break;
      
     case IMPULSE_MESSAGE_UBYTE:
      impulse_message_write_item(impulse_ubyte)
      break;
      
     case IMPULSE_MESSAGE_SHORT:
      impulse_message_write_item(impulse_short)
      break;
      
     case IMPULSE_MESSAGE_USHORT:
      impulse_message_write_item(impulse_ushort)
      break;
      
     case IMPULSE_MESSAGE_INT:
      impulse_message_write_item(impulse_int)
      break;
      
     case IMPULSE_MESSAGE_UINT:
      impulse_message_write_item(impulse_uint)
      break;
      
     case IMPULSE_MESSAGE_LONG:
      impulse_message_write_item(impulse_long)
      break;
      
     case IMPULSE_MESSAGE_ULONG:
      impulse_message_write_item(impulse_ulong)
      break;
      
     case IMPULSE_MESSAGE_FLOAT:
      impulse_message_write_item(impulse_float)
      break;
      
     case IMPULSE_MESSAGE_DOUBLE:
      impulse_message_write_item(impulse_double)
      break;
      
     case IMPULSE_MESSAGE_CHAR:
      impulse_message_write_item(impulse_char)
      break;
      
     case IMPULSE_MESSAGE_UNICHAR:
      impulse_message_write_item(impulse_unichar)
      break;
      
     case IMPULSE_MESSAGE_BOOL:
      impulse_message_write_item(impulse_bool)
      break;
      
     case IMPULSE_MESSAGE_LIST:
      {
	impulse_uint i;
	impulse_uint len = va_arg(args, impulse_uint);

	*((impulse_uint *)message->offset)++ = impulse_uint_be(len);
	
	switch (*++format)
	{
	 case IMPULSE_MESSAGE_BYTE:
	  impulse_message_write_list(impulse_byte)
	  break;
	  
	 case IMPULSE_MESSAGE_UBYTE:
	  impulse_message_write_list(impulse_ubyte)
	  break;
	  
	 case IMPULSE_MESSAGE_SHORT:
	  impulse_message_write_list(impulse_short)
	  break;
	  
	 case IMPULSE_MESSAGE_USHORT:
	  impulse_message_write_list(impulse_ushort)
	  break;
	  
	 case IMPULSE_MESSAGE_INT:
	  impulse_message_write_list(impulse_int)
	  break;
	  
	 case IMPULSE_MESSAGE_UINT:
	  impulse_message_write_list(impulse_uint)
	  break;
      
	 case IMPULSE_MESSAGE_LONG:
	  impulse_message_write_list(impulse_long)
	  break;
	  
	 case IMPULSE_MESSAGE_ULONG:
	  impulse_message_write_list(impulse_ulong)
	  break;
	  
	 case IMPULSE_MESSAGE_FLOAT:
	  impulse_message_write_list(impulse_float)
	  break;
	  
	 case IMPULSE_MESSAGE_DOUBLE:
	  impulse_message_write_list(impulse_double)
	  break;
	  
	 case IMPULSE_MESSAGE_CHAR:
	  impulse_message_write_list(impulse_char)
	  break;
	  
	 case IMPULSE_MESSAGE_UNICHAR:
	  impulse_message_write_list(impulse_unichar)
	  break;
	  
	 case IMPULSE_MESSAGE_BOOL:
	  impulse_message_write_list(impulse_bool)
	  break;

	 default:
	  {
	    impulse_uint i;
	    for (i = 0 ; i < nformatters ; i++)
	    {
	      if (formatters[i]->key == *format)
	      {
		formatters[i]->writer(message, len, va_arg(args, void *));
		break;
	      }
	    }
	  }
	  break;
	}
      }      
      break;
      
     default:
      {
	impulse_uint i;
	for (i = 0 ; i < nformatters ; i++)
	{
	  if (formatters[i]->key == *format)
	  {
	    formatters[i]->writer(message, 1, va_arg(args, void *));
	    break;
	  }
	}
      }
      /* Ohh this is bad.... but hey, we don't mind really */
      break;
    }
    format++;
  }
  va_end(args);
}


#define impulse_message_read_item(type) \
{ \
  type v = *((type *)message->offset); \
  ((type *)message->offset)++; \
  *va_arg(args, type *) = type##_be(v); \
}

#define impulse_message_read_list(type) \
{ \
  type *vs = impulse_alloc(sizeof(type)*len); \
  for (i = 0 ; i < len ; i++) \
  { \
    type v = *((type *)message->offset); \
    ((type *)message->offset)++; \
    vs[i] = type##_be(v); \
  } \
  *va_arg(args, type **) = vs; \
}


void impulse_message_readf(impulse_message message,
			   impulse_uint nformatters,
			   impulse_message_formatter *formatters,
			   impulse_char *format, 
			   ...)
{
  va_list args;
  va_start(args, format);
  
  while (*format)
  {
    switch (*format)
    {
     case IMPULSE_MESSAGE_BYTE:
      impulse_message_read_item(impulse_byte)
      break;
      
     case IMPULSE_MESSAGE_UBYTE:
      impulse_message_read_item(impulse_ubyte)
      break;
      
     case IMPULSE_MESSAGE_SHORT:
      impulse_message_read_item(impulse_short)
      break;
      
     case IMPULSE_MESSAGE_USHORT:
      impulse_message_read_item(impulse_ushort)
      break;
      
     case IMPULSE_MESSAGE_INT:
      impulse_message_read_item(impulse_int)
      break;
      
     case IMPULSE_MESSAGE_UINT:
      impulse_message_read_item(impulse_uint)
      break;
      
     case IMPULSE_MESSAGE_LONG:
      impulse_message_read_item(impulse_long)
      break;
      
     case IMPULSE_MESSAGE_ULONG:
      impulse_message_read_item(impulse_ulong)
      break;
      
     case IMPULSE_MESSAGE_FLOAT:
      impulse_message_read_item(impulse_float)
      break;
      
     case IMPULSE_MESSAGE_DOUBLE:
      impulse_message_read_item(impulse_double)
      break;
      
     case IMPULSE_MESSAGE_CHAR:
      impulse_message_read_item(impulse_char)
      break;
      
     case IMPULSE_MESSAGE_UNICHAR:
      impulse_message_read_item(impulse_unichar)
      break;
      
     case IMPULSE_MESSAGE_BOOL:
      impulse_message_read_item(impulse_bool)
      break;
      
     case IMPULSE_MESSAGE_LIST:
      {
	impulse_uint i;
	impulse_uint len = *((impulse_uint *)message->offset);
	((impulse_uint *)message->offset)++;
	*va_arg(args, impulse_uint *) = impulse_uint_be(len);

	len = impulse_uint_be(len);

	switch (*++format)
	{
	 case IMPULSE_MESSAGE_BYTE:
	  impulse_message_read_list(impulse_byte)
	  break;
	  
	 case IMPULSE_MESSAGE_UBYTE:
	  impulse_message_read_list(impulse_ubyte)
	  break;
	  
	 case IMPULSE_MESSAGE_SHORT:
	  impulse_message_read_list(impulse_short)
	  break;
	  
	 case IMPULSE_MESSAGE_USHORT:
	  impulse_message_read_list(impulse_ushort)
	  break;
	  
	 case IMPULSE_MESSAGE_INT:
	  impulse_message_read_list(impulse_int)
	  break;
	  
	 case IMPULSE_MESSAGE_UINT:
	  impulse_message_read_list(impulse_uint)
	  break;
      
	 case IMPULSE_MESSAGE_LONG:
	  impulse_message_read_list(impulse_long)
	  break;
	  
	 case IMPULSE_MESSAGE_ULONG:
	  impulse_message_read_list(impulse_ulong)
	  break;
	  
	 case IMPULSE_MESSAGE_FLOAT:
	  impulse_message_read_list(impulse_float)
	  break;
	  
	 case IMPULSE_MESSAGE_DOUBLE:
	  impulse_message_read_list(impulse_double)
	  break;
	  
	 case IMPULSE_MESSAGE_CHAR:
	  impulse_message_read_list(impulse_char)
	  break;
	  
	 case IMPULSE_MESSAGE_UNICHAR:
	  impulse_message_read_list(impulse_unichar)
	  break;
	  
	 case IMPULSE_MESSAGE_BOOL:
	  impulse_message_read_list(impulse_bool)
	  break;
	  
	 default:
	  {
	    impulse_uint i;
	    for (i = 0 ; i < nformatters ; i++)
	    {
	      if (formatters[i]->key == *format)
	      {
		formatters[i]->reader(message, len, va_arg(args, void *));
		break;
	      }
	    }
	  }
	  break;
	}
      }      
      break;

     default:
      {
	impulse_uint i;
	for (i = 0 ; i < nformatters ; i++)
	{
	  if (formatters[i]->key == *format)
	  {
	    formatters[i]->reader(message, 1, va_arg(args, void *));
	    break;
	  }
	}
      }
      /* Ohh this is bad.... but hey, we don't mind really */
      break;
    }
    format++;
  }
  
  va_end(args);
}


/* End of file */
