/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * client.c
 * --------
 *
 * Communication transport functions
 * Client implementation
 */

#include "../../../config.h"

#include <impulse/comms/transport.h>
#include <impulse/comms/transport_sock.h>
#include <impulse/util/memory.h>
#include <impulse/util/thread.h>
#include <impulse/util/hash.h>


#if USE_THREADS
static impulse_mutex_static transport_lock = IMPULSE_MUTEX_CREATE_STATIC;
#endif
static impulse_hash transport_map = NULL;


static void impulse_transport_init(void)
{
  if (!transport_map)
  {
    impulse_int i;
    
    transport_map = impulse_hash_create();
    
#if HAVE_COMMS_SOCK
    impulse_transport_module_sock->init();
    
    for (i = 0 ; i < impulse_transport_module_sock->nprotocol ; i++)
      impulse_hash_store(transport_map, impulse_transport_module_sock->protocol[i], impulse_transport_module_sock);
#endif
  }
}


void impulse_transport_register(impulse_transport_module module)
{
  impulse_int i;
#if USE_THREADS
  impulse_mutex_lock(&transport_lock);
#endif

  if (!transport_map)
    impulse_transport_init();
  
  module->init();
  
  for (i = 0 ; i < module->nprotocol ; i++)
    impulse_hash_store(transport_map, module->protocol[i], module);
  
#if USE_THREADS
  impulse_mutex_unlock(&transport_lock);
#endif
}


void impulse_transport_deregister(impulse_transport_module module)
{
  impulse_int i;
#if USE_THREADS
  impulse_mutex_lock(&transport_lock);
#endif
  
  if (!transport_map)
    impulse_transport_init();

  for (i = 0 ; i < module->nprotocol ; i++)
    impulse_hash_remove(transport_map, module->protocol[i]);
  module->cleanup();
  
#if USE_THREADS
  impulse_mutex_unlock(&transport_lock);
#endif
}


impulse_transport_impl impulse_transport_load(impulse_unistring protocol)
{
  impulse_transport_module module = NULL;
#if USE_THREADS
  impulse_mutex_lock(&transport_lock);
#endif
  
  if (!transport_map)
    impulse_transport_init();

  module = impulse_hash_recover(transport_map, protocol);
  
#if USE_THREADS
  impulse_mutex_unlock(&transport_lock);
#endif
  
  return module->impl;
}


void impulse_transport_unload(impulse_unistring protocol)
{
#if USE_THREADS
  impulse_mutex_lock(&transport_lock);
#endif
  
  if (!transport_map)
    impulse_transport_init();

#if USE_THREADS
  impulse_mutex_unlock(&transport_lock);
#endif
}


impulse_transport impulse_transport_create(impulse_unistring protocol)
{
  impulse_transport transport = NULL;
  impulse_transport_impl impl = NULL;
  
  if (!(impl = impulse_transport_load(protocol)))
    return NULL;
  
  transport = impulse_alloc(sizeof(struct _impulse_transport));
  transport->impl = impl;
  transport->protocol = impulse_unistring_create(protocol);
  transport->internal = NULL;
#if USE_THREADS
  transport->lock = impulse_mutex_create_recursive();
#endif
  
  transport->impl->create(transport);
  
  return transport;
}


void impulse_transport_destroy(impulse_transport transport)
{
  transport->impl->destroy(transport);
#if USE_THREADS
  impulse_mutex_destroy(transport->lock);
#endif
  impulse_unistring_destroy(transport->protocol);
  impulse_free(transport);
}


/* End of file */
