/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport_pipe.c
 * ----------------
 *
 * Communication transport functions
 * pipeets protocols
 * 
 * Addresses are specified in the form:
 * 
 * protocol:node:service
 * 
 * For IP based services, 'node' can be a host name,
 * IPv4 dotted decimal address, or IPv6 colon separated
 * hexidecimal address. 'service' is either a numeric
 * port number or service name from /etc/services or the like
 * 
 * For Unix domain pipeets, 'node' must be one of 'unix' or
 * 'local', and 'service' is a full path name.
 * 
 * eg.
 * 
 * pipe:www.google.com:http
 * pipe:10.0.0.1:ftp
 * pipe:localhost:6000
 * pipe:unix:/var/run/mypipeet
 * pipe:f340:f353::f454:telnet
 * 
 */

#include "../../../config.h"

#include "impulse/comms/transport_pipe_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/error.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>


/*                            ======--------======                           */


/* Static unicode strings for protocol name */
static struct _impulse_unistring _name_pipe =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'p',
    (impulse_unichar)'i',
    (impulse_unichar)'p', 
    (impulse_unichar)'e',
    (impulse_unichar)'\0'
  }
};


/* Transport function blocks */

static struct _impulse_transport_impl _pipe_impl =
{
  impulse_transport_pipe_create,
  impulse_transport_pipe_destroy,

  /* Event functions */
  impulse_transport_pipe_handle,
  impulse_transport_pipe_unhandle,
  
  /* Name resolution */
  impulse_transport_pipe_lookup,
  impulse_transport_pipe_rlookup,
  
  /* Connection functions */
  impulse_transport_pipe_open,
  impulse_transport_pipe_accept,
  impulse_transport_pipe_close,
  impulse_transport_pipe_active,

  /* Data transfer functions */
  impulse_transport_pipe_pending,
  impulse_transport_pipe_send,
  impulse_transport_pipe_receive,
};


struct _impulse_transport_module _impulse_transport_module_pipe = {
  &_pipe_impl,
  1,
  (impulse_unistring[]) {
    &_name_pipe
  },
  impulse_transport_pipe_init,
  impulse_transport_pipe_cleanup
};

impulse_transport_module impulse_transport_module_pipe = &_impulse_transport_module_pipe;


/*                            ======--------======                           */


/* Evnet helper functions */

static impulse_bool impulse_transport_pipe_event(impulse_event *event,
						 void *data)
{
  if (event->type == IMPULSE_EVENT_FILEDESCRIPTOR &&
      event->data.fd.type == IMPULSE_EVENT_FD_READ)
  {
    impulse_transport transport = (impulse_transport)data;
    impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
    
    if (pipe->handler)
      return (*pipe->handler)(transport,
			      IMPULSE_TRANSPORT_EVENT_DATA,
			      pipe->edata);
  }
  return 0;
}


/* Name parsing helper function */

static impulse_int impulse_transport_pipe_parse(impulse_unistring address,
						impulse_int *handle)
{
  impulse_unistring node;
  impulse_int offset;
  impulse_char *start, *end = NULL;
  
  offset = impulse_unistring_find_chr(address, (impulse_unichar)':');
  
  if (offset < 0)
    return -1;
  
  node = impulse_unistring_create_space(address->length - (offset+1));
  impulse_unistring_substring(node, address, offset+1, address->length);
  start = impulse_unistring_to_utf(node,8,0);
  impulse_unistring_destroy(node);
  
  *handle = strtoul(start, &end, 10);
  impulse_free(start);

  if (start == end)
    return -1;
  
  return 0;
}

/*                            ======--------======                           */

void impulse_transport_pipe_init(void)
{
}

void impulse_transport_pipe_cleanup(void)
{
}


void impulse_transport_pipe_create(impulse_transport transport)
{
  impulse_transport_pipe pipe = impulse_alloc(sizeof(struct _impulse_transport_pipe));
  
  pipe->in = -1;
  pipe->out = -1;
  
  pipe->handler = NULL;
  pipe->edata = NULL;
  
  transport->internal = pipe;
}


void impulse_transport_pipe_destroy(impulse_transport transport)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  
  if (pipe->in != -1)
    impulse_transport_pipe_close(transport);
  
  impulse_free(pipe);
}


/* Event functions */
impulse_int impulse_transport_pipe_handle(impulse_transport transport,
					  impulse_transport_handler func,
					  void *edata)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  
  pipe->handler = func;
  pipe->edata = edata;
  
  return 0;
}


impulse_int impulse_transport_pipe_unhandle(impulse_transport transport)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  
  if (pipe->handler)
  {
    pipe->handler = NULL;
    pipe->edata = NULL;
  }
  
  return 0;
}


impulse_int impulse_transport_pipe_lookup(impulse_transport transport,
					  impulse_unistring name,
					  impulse_unistring *address)
{
  *address = impulse_unistring_create(name);
  return 0;
}


impulse_int impulse_transport_pipe_rlookup(impulse_transport transport,
					   impulse_unistring address,
					   impulse_unistring *name)
{
  *name = impulse_unistring_create(address);
  return 0;
}


/* Connection functions */
impulse_int impulse_transport_pipe_open(impulse_transport transport,
					impulse_unistring local,
					impulse_unistring remote)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;

  if (!local)
    return -1;
  if (!remote)
    return -1;
  if (pipe->in != -1)
    return -1;

  if (impulse_transport_pipe_parse(local,
				   &pipe->in) < 0)
    return -1;
  if (impulse_transport_pipe_parse(remote,
				   &pipe->out) < 0)
  {
    pipe->in = -1;
    return -1;
  }

  transport->local = impulse_unistring_create(local);
  transport->remote = impulse_unistring_create(remote);

  impulse_event_register_fd(pipe->out,
			    IMPULSE_EVENT_FD_READ,
			    impulse_transport_pipe_event,
			    NULL,
			    transport);
  return 0;
}


impulse_int impulse_transport_pipe_accept(impulse_transport transport,
					  impulse_transport *client)
{
  /* Pipes are peer-to-peer, so no need to accept */
  return -1;
}


impulse_int impulse_transport_pipe_close(impulse_transport transport)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  
  if (pipe->in != -1)
  {
    impulse_event_deregister_fd(pipe->in);
    impulse_event_deregister_fd(pipe->out);
    close(pipe->in);
    close(pipe->out);
    pipe->in = -1;
    pipe->out = -1;
    
    if (transport->local)
      impulse_unistring_destroy(transport->local);
    if (transport->remote)
      impulse_unistring_destroy(transport->remote);
    
    transport->local = NULL;
    transport->remote = NULL;
  }
  return 0;
}


impulse_int impulse_transport_pipe_active(impulse_transport transport)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;

  return pipe->in != -1;
}


/* Data transfer functions */
impulse_int impulse_transport_pipe_pending(impulse_transport transport)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  impulse_int pending;
  impulse_int ret;

  if (pipe->in == -1)
    return -1;
  
  ret = ioctl(pipe->in, FIONREAD, &pending);

  if (ret == -1)
    return -1;
  return pending;
}


impulse_int impulse_transport_pipe_send(impulse_transport transport,
					impulse_byte *buffer,
					impulse_uint size,
					impulse_unistring address)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  impulse_int ret = 0;

  if (pipe->in == -1)
    return -1;

  if ((ret = write(pipe->out, (char *)buffer, size)) < 0)
    return -1;

  return 0;
}


impulse_int impulse_transport_pipe_receive(impulse_transport transport,
					   impulse_byte *buffer,
					   impulse_uint size,
					   impulse_unistring *address)
{
  impulse_transport_pipe pipe = (impulse_transport_pipe)transport->internal;
  impulse_int ret = 0;
  
  if (pipe->in == -1)
    return -1;
  
  if ((ret = read(pipe->in, (char *)buffer, size)) == -1)
    return -1;

  return ret;
}


/* End of file */
