/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * transport_sock.c
 * ----------------
 *
 * Communication transport functions
 * Sockets protocols
 * 
 * Addresses are specified in the form:
 * 
 * protocol:node:service
 * 
 * For IP based services, 'node' can be a host name,
 * IPv4 dotted decimal address, or IPv6 colon separated
 * hexidecimal address. 'service' is either a numeric
 * port number or service name from /etc/services or the like
 * 
 * For Unix domain sockets, 'node' must be one of 'unix' or
 * 'local', and 'service' is a full path name.
 * 
 * eg.
 * 
 * sock:www.google.com:www
 * sock:10.0.0.1:ftp
 * sock:localhost:6000
 * sock:unix:/var/run/mysocket
 * sock:f340:f353::f454:telnet
 * 
 */

#include "../../../config.h"

#include "impulse/comms/transport_sock_int.h"

#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/hash.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include <signal.h>

#if HAVE_COMMS_SOCK

#include <netdb.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <arpa/inet.h>

#include <sys/wait.h>

/*                            ======--------======                           */



union sockaddr_all
{
  struct sockaddr a;
#if HAVE_COMMS_SOCK_UNIX || HAVE_COMMS_SOCK_LOCAL
  struct sockaddr_un un;
#endif
#if HAVE_COMMS_SOCK_IPV4
  struct sockaddr_in in;
#endif
#if HAVE_COMMS_SOCK_IPV6
  struct sockaddr_in6 in6;
#endif
};


typedef struct _impulse_transport_sock_info
{
  impulse_int family;
  impulse_byte protocol;
  impulse_int type;
  impulse_unistring name;
} impulse_transport_sock_info;


/*                            ======--------======                           */


/* Static unicode strings for protocol name */
static struct _impulse_unistring _name_tcp4 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p',
    (impulse_unichar)'4',
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_tcp6 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'6', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_tcpl =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'t',
    (impulse_unichar)'c',
    (impulse_unichar)'p', 
    (impulse_unichar)'l',
    (impulse_unichar)'\0'
  }
};

static struct _impulse_unistring _name_udp4 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p',
    (impulse_unichar)'4',
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udp6 =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'6', 
    (impulse_unichar)'\0'
  }
};
static struct _impulse_unistring _name_udpl =
{
  4, 4,
  (impulse_unichar[]){
    (impulse_unichar)'u',
    (impulse_unichar)'d',
    (impulse_unichar)'p', 
    (impulse_unichar)'l',
    (impulse_unichar)'\0'
  }
};

/* Transport function blocks */

static struct _impulse_transport_impl _sock_impl =
{
  impulse_transport_sock_create,
  impulse_transport_sock_destroy,

  /* Event functions */
  impulse_transport_sock_handle,
  impulse_transport_sock_unhandle,
  
  /* Name resolution stuff */
  impulse_transport_sock_lookup,
  impulse_transport_sock_rlookup,
  
  /* Connection functions */
  impulse_transport_sock_open,
  impulse_transport_sock_accept,
  impulse_transport_sock_close,
  impulse_transport_sock_active,
  
  /* Data transfer functions */
  impulse_transport_sock_pending,
  impulse_transport_sock_send,
  impulse_transport_sock_receive,
};

static struct _impulse_transport_module _impulse_transport_module_sock = { 
  &_sock_impl, 
  6,
  (impulse_unistring[]){
    &_name_tcp4, &_name_tcp6, &_name_tcpl, 
    &_name_udp4, &_name_udp6, &_name_udpl
  },
  impulse_transport_sock_init,
  impulse_transport_sock_cleanup
};

impulse_transport_module impulse_transport_module_sock = &_impulse_transport_module_sock;

impulse_transport_sock_info sock_info[IMPULSE_TRANSPORT_SOCK_MAX] = 
{
#if HAVE_COMMS_SOCK_IPV4
  { AF_INET, PF_INET, SOCK_STREAM, &_name_tcp4},
  { AF_INET, PF_INET, SOCK_DGRAM, &_name_udp4},
#else
  { 0, 0, 0, NULL},
  { 0, 0, 0, NULL},
#endif
  
#if HAVE_COMMS_SOCK_IPV6
  { AF_INET6, PF_INET6, SOCK_STREAM, &_name_tcp6},
  { AF_INET6, PF_INET6, SOCK_DGRAM, &_name_udp6},
#else
  { 0, 0, 0, NULL},
  { 0, 0, 0, NULL},
#endif
  
#if HAVE_COMMS_SOCK_LOCAL
  { AF_LOCAL, PF_LOCAL, SOCK_STREAM, &_name_tcpl},
  { AF_LOCAL, PF_LOCAL, SOCK_DGRAM, &_name_udpl}
#elif HAVE_COMMS_SOCK_UNIX
  { AF_UNIX, PF_UNIX, SOCK_STREAM, &_name_tcpl},
  { AF_UNIX, PF_UNIX, SOCK_DGRAM, &_name_udpl}
#else
  { 0, 0, 0, NULL},
  { 0, 0, 0, NULL},
#endif
};

/* Name server stuff */

typedef struct _impulse_transport_sock_dns
{
  impulse_bool inprogress:1;
  union
  {
    struct
    {
      impulse_int status;
      impulse_unistring address;
    } entry;
    
    struct
    {
      impulse_int pid;
      impulse_int handle;
    } resolver;
  } data;
} *impulse_transport_sock_dns;


static impulse_hash name_cache = NULL;
static impulse_hash rname_cache = NULL;

/*                            ======--------======                           */


/* Event helper functions */

static impulse_bool impulse_transport_sock_client(impulse_event *event,
						  void *data)
{
  impulse_bool ret = 0;
  impulse_transport transport = (impulse_transport)data;
  impulse_transport_lock(transport);

  if (event->type == IMPULSE_EVENT_FILEDESCRIPTOR)
  {
    if (event->data.fd.type == IMPULSE_EVENT_FD_READ)
    {
      /* We've got some date */
      impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
      
      if (sock->handler)
	ret = (*sock->handler)(transport,
			       IMPULSE_TRANSPORT_EVENT_DATA,
			       sock->edata);
    }
    else if (event->data.fd.type == IMPULSE_EVENT_FD_WRITE)
    {
      /* This just tells us the socket is connected */
      impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
      
      if (sock->handle)
	ret = (*sock->handler)(transport,
			       IMPULSE_TRANSPORT_EVENT_CONNECT,
			       sock->edata);
      
      /* We 're now interested in receiving data */
      impulse_event_deregister_fd(sock->handle);
      impulse_event_register_fd(sock->handle,
				IMPULSE_EVENT_FD_READ,
				impulse_transport_sock_client,
				NULL,
				data);
    }
    else
    {
      printf("Exception\n");
    }
  }
  impulse_transport_unlock(transport);
  return ret;
}


static impulse_bool impulse_transport_sock_server(impulse_event *event,
						  void *data)
{
  impulse_bool ret = 0;
  impulse_transport transport = (impulse_transport)data;
  impulse_transport_lock(transport);
  
  if (event->type == IMPULSE_EVENT_FILEDESCRIPTOR &&
      event->data.fd.type == IMPULSE_EVENT_FD_READ)
  {
    /* We've got an incoming connection request */
    impulse_transport transport = (impulse_transport)data;
    impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
    
    if (sock->handler)
      ret = (*sock->handler)(transport,
			     IMPULSE_TRANSPORT_EVENT_INCOMING,
			     sock->edata);
  }
  impulse_transport_unlock(transport);
  return ret;
}

static impulse_bool impulse_transport_sock_resolver(impulse_event *event,
						    void *data)
{
  impulse_bool ret = 0;
  impulse_transport transport = (impulse_transport)data;
  impulse_transport_lock(transport);

  if (event->type == IMPULSE_EVENT_FILEDESCRIPTOR &&
      event->data.fd.type == IMPULSE_EVENT_FD_READ)
  {
    /* We've got an incoming connection request */
    impulse_transport transport = (impulse_transport)data;
    impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
    
    if (sock->handler)
      ret = (*sock->handler)(transport,
			     IMPULSE_TRANSPORT_EVENT_RESOLVE,
			     sock->edata);
  }
  impulse_transport_unlock(transport);
  return ret;
}


/* Address helper functions */


static impulse_int impulse_transport_sock_parse(impulse_byte protocol,
						impulse_unistring address,
						union sockaddr_all **addr,
						impulse_int *addr_len,
						impulse_bool passive)
{
  impulse_unistring unode;
  impulse_unistring uservice;
  impulse_char *node = NULL;
  impulse_char *service = NULL;
  impulse_int offset;
  impulse_int ret = 0;

  /*
   * First of all split the address into a protocol,node,service
   * triplet
   */
  offset = impulse_unistring_rfind_chr(address, (impulse_unichar)':');
  if (offset < 0)
    return IMPULSE_TRANSPORT_ADDRESS_INVALID;

  if ((offset) > 0)
  {
    unode = impulse_unistring_create_sub(address, 0, offset);
    node = impulse_unistring_to_utf(unode, 8, 0);
    impulse_unistring_destroy(unode);
  }

  if ((address->length - offset) > 1)
  {
    uservice = impulse_unistring_create_sub(address, offset+1, address->length);
    service = impulse_unistring_to_utf(uservice, 8, 0);
    impulse_unistring_destroy(uservice);
  }

  /*
   * Now parse the protocol and service parts
   */
  switch (protocol)
  {
#if HAVE_COMMS_SOCK_LOCAL || HAVE_COMMS_SOCK_UNIX
    /*
     * For local domain sockets, the node should be
     * one of 'local' or 'unix', and the service should
     * be the fully qualified path name NB. path must be
     * on a local volume. ie. NOT nfs
     */
   case IMPULSE_TRANSPORT_SOCK_TCPL:
   case IMPULSE_TRANSPORT_SOCK_UDPL:
    if (!strcmp(node, "local") ||
	!strcmp(node, "unix"))
    {
      (*addr) = impulse_alloc(sizeof(union sockaddr_all));
      (*addr)->un.sun_family = sock_info[protocol].family;
      strncpy((*addr)->un.sun_path, service, 107);
      (*addr)->un.sun_path[107] = '\0';
      *addr_len = SUN_LEN(&(*addr)->un);
      ret = 0;
    }
    else
      ret = IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;
    break;
#endif

#if HAVE_COMMS_SOCK_IPV4 || HAVE_COMMS_SOCK_IPV6
    /*
     * For IP4/6 based sockets the node must be the
     * raw ip address or dns name , and service can be a port number
     * or service from /etc/services.
     * If a domain name is specified, this function may
     * block indefinitely (well for several minutes at worst)
     */
   case IMPULSE_TRANSPORT_SOCK_TCP4:
   case IMPULSE_TRANSPORT_SOCK_UDP4:
   case IMPULSE_TRANSPORT_SOCK_TCP6:
   case IMPULSE_TRANSPORT_SOCK_UDP6:
    {
      impulse_int port = 0;

      /*
       * Fix me: replace with reentrant ones if available
       * or make this stuff thread safe with mutexes
       */
      
      /*
       * First off get hold of the service 
       */
      if (service)
      {
	struct servent *se = NULL;
	impulse_char *end = NULL;
	port = strtoul(service, &end, 10);
	if (end == service) /* service is not numeric */
	{
	  if (!(se = getservbyname(service, (sock_info[protocol].type == SOCK_STREAM)?"tcp" : "udp")))
	  {
	    ret = IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;
	    break;
	  }
	  port = se->s_port;
	}
	else
	  port = htons(port);
      }
      else /* We must have a service */
      {
	ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
	break;
      }

      switch (sock_info[protocol].family)
      {
# if HAVE_COMMS_SOCK_IPV4
       case AF_INET:
	{
	  struct in_addr bin;

	  if (!node)
	    bin.s_addr = htonl(passive ? INADDR_ANY : INADDR_LOOPBACK);
	  else if (!inet_aton(node, &bin))
	    ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
	  
	  if (ret == 0)
	  {
	    (*addr) = impulse_alloc(sizeof(union sockaddr_all));
	    *addr_len = sizeof(union sockaddr_all);
	    
	    (*addr)->in.sin_family = AF_INET;
	    (*addr)->in.sin_port = port;
	    (*addr)->in.sin_addr = bin;
	    memset(&((*addr)->in.sin_zero), 0, 8);
	  }
	}	
	break;
# endif
# if HAVE_COMMS_SOCK_IPV6
       case AF_INET6:
	{
	  struct in6_addr bin;

	  if (!node)
	    bin = (passive ? in6addr_any : in6addr_loopback);
	  else if (!inet_pton(AF_INET6, node, &bin))
	    ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
	  
	  if (ret == 0)
	  {
	    (*addr) = impulse_alloc(sizeof(union sockaddr_all));
	    *addr_len = sizeof(union sockaddr_all);
	    
	    (*addr)->in6.sin6_family = AF_INET6;
	    (*addr)->in6.sin6_port = port;
	    (*addr)->in6.sin6_flowinfo = 0;
	    (*addr)->in6.sin6_addr = bin;
	  }
	}
	break;
# endif
       default:
	ret = IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;
	break;
      }
    }
    break;
#endif
    
   default:
    ret = IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;
    break;
  }

  return ret;
}


static impulse_int impulse_transport_sock_format(impulse_byte protocol,
						 union sockaddr_all *addr,
						 impulse_int addr_len,
						 impulse_unistring *address)
{
  impulse_int ret = 0;
  
  switch (addr->a.sa_family)
  {
#if HAVE_COMMS_SOCK_LOCAL
   case AF_LOCAL:
#elif HAVE_COMMS_SOCK_UNIX
   case AF_UNIX:
#endif
#if HAVE_COMMS_SOCK_LOCAL || HAVE_COMMS_SOCK_UNIX
    {
      *address = impulse_unistring_create_utf("local:", 8);
      impulse_unistring_append_utf(*address, addr->un.sun_path, 8);
    }
    break;
#endif

#if HAVE_COMMS_SOCK_IPV4
   case AF_INET:
    {
      impulse_char *node = NULL;
      impulse_short service;
      impulse_char service_buf[50];
      
      if (!(node = inet_ntoa(addr->in.sin_addr)))
      {
	ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
	break;
      }

      service = ntohs(addr->in.sin_port);
      sprintf(service_buf, "%d", service);

      *address = impulse_unistring_create_space(strlen(node)+strlen(service_buf)+1);
      impulse_unistring_append_utf(*address, node, 8);
      impulse_unistring_append_utf(*address, ":", 8);
      impulse_unistring_append_utf(*address, service_buf, 8);
    }
    break;
#endif
    
#if HAVE_COMMS_SOCK_IPV6
   case AF_INET6:
    {
      impulse_short service;
      impulse_char service_buf[50];
      impulse_char node_buf[100];

      if (!inet_ntop(AF_INET6, &addr->in6.sin6_addr, node_buf, 100))
      {
	ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
	break;
      }
      service = ntohs(addr->in6.sin6_port);
      sprintf(service_buf, "%d", service);

      *address = impulse_unistring_create_space(strlen(node_buf)+strlen(service_buf)+1);
      impulse_unistring_append_utf(*address, node_buf, 8);
      impulse_unistring_append_utf(*address, ":", 8);
      impulse_unistring_append_utf(*address, service_buf, 8);
    }
    break;
#endif

   default:
    ret = IMPULSE_TRANSPORT_ADDRESS_INVALID;
    break;
  }
  
  return ret;
}


/* Name resolution stuff */

/*
 * Does a blocking Host Name <-> raw ip address conversion
 */
static impulse_int impulse_transport_sock_resolve(impulse_int family,
						  impulse_char *in,
						  impulse_char **out,
						  impulse_bool reverse)
{
  struct hostent *he;
  char buf[500];

  if (!reverse)
  {
    if (!(he = gethostbyname2(in, family)))
      return IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;

    if (!inet_ntop(family, he->h_addr, buf, 499))
      return IMPULSE_TRANSPORT_ADDRESS_INVALID;

    *out = impulse_alloc(strlen(buf)+1);
    strcpy(*out,buf);
  }
  else
  {
    if (!inet_pton(family, in, buf))
      return IMPULSE_TRANSPORT_ADDRESS_INVALID;

    if (!(he = gethostbyaddr(buf, (family == AF_INET ? 4 : 16), family)))
      return IMPULSE_TRANSPORT_ADDRESS_UNKNOWN;

    *out = impulse_alloc(strlen(he->h_name)+1);
    strcpy(*out, he->h_name);
  }

  return 0;
}


static void impulse_transport_sock_dns_start(impulse_transport transport,
					     impulse_int family,
					     impulse_unistring node,
					     impulse_transport_sock_dns *record,
					     impulse_bool reverse)
{
  impulse_int pid;
  impulse_int handle[2];
  
  if (pipe(handle) < 0)
    ;
  
  pid = fork();
  
  switch (pid)
  {
   case 0: /* Child */
    {
      impulse_char *in = NULL;
      impulse_char *out = NULL;
      impulse_int status = 0;
      close(handle[0]);
      
      /* 
       * NB, I can't be bothered to free any memory here, cos we
       * will be exiting in about 10 lines time
       */
      in = impulse_unistring_to_utf(node,8,0);
      status = impulse_transport_sock_resolve(family, in, &out, reverse);

      write(handle[1], &status, sizeof(status));
      if (status >= 0)
      {
	impulse_unistring result = impulse_unistring_create_utf(out,8);
	write(handle[1], &result->length, sizeof(result->length));
	write(handle[1], result->data, sizeof(impulse_unichar) * result->length);
      }
      _exit(0);
    }
    break;
    
   case -1:
    /* Fix me: we should probably fall back on a blocking dns lookup */
    break;
    
   default:
    close(handle[1]);
    *record = impulse_alloc(sizeof(struct _impulse_transport_sock_dns));
    (*record)->inprogress = 1;
    (*record)->data.resolver.pid = pid;
    (*record)->data.resolver.handle = handle[0];
    impulse_event_register_fd(handle[0],
			      IMPULSE_EVENT_FD_READ,
			      impulse_transport_sock_resolver,
			      NULL,
			      transport);
    break;
  }
}


static void impulse_transport_sock_dns_end(impulse_transport_sock_dns record)
{
  struct timeval tv = { 0,0 };
  fd_set fs;
  impulse_int handle = record->data.resolver.handle;
  
  FD_ZERO(&fs);
  FD_SET(handle, &fs);
  if (select(handle+1, &fs, NULL, NULL, &tv) < 0)
  {
  }
  if (FD_ISSET(handle, &fs))
  {
    impulse_int status;
    impulse_int length;
    impulse_unistring name = NULL;
    
    read(handle, &status, sizeof(status));
    if (status >= 0)
    {
      read(handle, &length, sizeof(length));
      name = impulse_unistring_create_space(length);
      read(handle, name->data, sizeof(impulse_unichar)*length);
      name->length = length;
      name->data[length] = 0;
    }
    impulse_event_deregister_fd(handle);
    close(handle);
    waitpid(record->data.resolver.pid, NULL, 0);
    
    record->inprogress = 0;
    record->data.entry.status = status;
    record->data.entry.address = name;
  }
}



/*                            ======--------======                           */

static void resolver_cleaner(void *data)
{
  impulse_transport_sock_dns resolver = (impulse_transport_sock_dns)data;
  
  if (resolver->inprogress)
  {
    kill(resolver->data.resolver.pid, SIGTERM);
    waitpid(resolver->data.resolver.pid, NULL, WNOHANG);
    close(resolver->data.resolver.handle);
  }
  else
    impulse_unistring_destroy(resolver->data.entry.address);
    
  impulse_free(resolver);
}


void impulse_transport_sock_init(void)
{
  name_cache = impulse_hash_create();
  rname_cache = impulse_hash_create();
  
  impulse_hash_cleaner(name_cache,
		       resolver_cleaner);
  impulse_hash_cleaner(rname_cache,
		       resolver_cleaner);
}


void impulse_transport_sock_cleanup(void)
{
  impulse_hash_destroy(name_cache);
  impulse_hash_destroy(rname_cache);
}


void impulse_transport_sock_create(impulse_transport transport)
{
  impulse_transport_sock sock = impulse_alloc(sizeof(struct _impulse_transport_sock));
  impulse_int i;
  
  sock->handle = -1;
  sock->protocol = -1;
  
  sock->handler = NULL;
  sock->edata = NULL;
  
  transport->internal = sock;

  for (i = 0 ; i < IMPULSE_TRANSPORT_SOCK_MAX ; i++)
  {
    if (sock_info[i].name &&
	(impulse_unistring_starts(transport->protocol, sock_info[i].name)))
    {
      sock->protocol = i;
      break;
    }
  }
  if (sock->protocol == -1)
    ; /* Feck arse */
}


void impulse_transport_sock_destroy(impulse_transport transport)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  
  if (sock->handle != -1)
    impulse_transport_sock_close(transport);
  
  impulse_free(sock);
}


/* Event functions */
impulse_int impulse_transport_sock_handle(impulse_transport transport,
					  impulse_transport_handler func,
					  void *edata)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_transport_lock(transport);
  
  if (sock->handler)
    impulse_transport_sock_unhandle(transport);

  sock->handler = func;
  sock->edata = edata;

  impulse_transport_unlock(transport);
  return 0;
}


impulse_int impulse_transport_sock_unhandle(impulse_transport transport)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_transport_lock(transport);

  if (sock->handler)
  {
    sock->handler = NULL;
    sock->edata = NULL;
  }
  
  impulse_transport_unlock(transport);
  return 0;
}


/* Name resolution stuff */

impulse_int impulse_transport_sock_lookup(impulse_transport transport,
					  impulse_unistring name,
                                          impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_transport_sock_dns record = NULL;
  impulse_int offset = impulse_unistring_rfind_chr(name, (impulse_unichar)':');
  impulse_int ret = 0;
  impulse_unistring key;

  if (offset < 0)
    return IMPULSE_TRANSPORT_ADDRESS_INVALID;

  impulse_transport_lock(transport);

  /* 
   * Short circuit for local domain sockets and where
   * node is empty
   */
  if (sock->protocol == IMPULSE_TRANSPORT_SOCK_TCPL ||
      sock->protocol == IMPULSE_TRANSPORT_SOCK_UDPL ||
      offset == 0)
  {
    *address = impulse_unistring_create(name);
    impulse_transport_unlock(transport);
    return 0;
  }
  
  key = impulse_unistring_create_sub(name, 0, offset);
  impulse_unistring_prepend_utf(key, ":", 8);
  impulse_unistring_prepend(key, transport->protocol);
  
  if (!(record = impulse_hash_recover(name_cache, key))) /* There is no cached entry */
  {
    impulse_unistring node = impulse_unistring_create_sub(name, 0, offset);
    impulse_transport_sock_dns_start(transport, sock_info[sock->protocol].family, node, &record, 0);
    impulse_hash_store(name_cache, key, record);
    impulse_unistring_destroy(node);
    ret = IMPULSE_TRANSPORT_ADDRESS_RESOLVE;
  }
  else
  {
    if (record->inprogress) /* The cached entry is in progress */
      impulse_transport_sock_dns_end(record);
    
    if (!record->inprogress) /* The name lookup is complete */
    {
      if (record->data.entry.status < 0)
	ret = record->data.entry.status;
      else
      {
	*address = impulse_unistring_create_sub(name, offset, name->length);
	impulse_unistring_prepend(*address, record->data.entry.address);
      }
    }
    else
      ret = IMPULSE_TRANSPORT_ADDRESS_RESOLVE;
  }
  
  impulse_unistring_destroy(key);
  
  impulse_transport_unlock(transport);
  return ret;
}


impulse_int impulse_transport_sock_rlookup(impulse_transport transport,
					   impulse_unistring address,
                                           impulse_unistring *name)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_transport_sock_dns record = NULL;
  impulse_int offset = impulse_unistring_rfind_chr(address, (impulse_unichar)':');
  impulse_int ret = 0;
  impulse_int proto;
  impulse_unistring key;
  
  if (offset < 0)
    return IMPULSE_TRANSPORT_ADDRESS_INVALID;

  impulse_transport_lock(transport);
  /* Short circuit for local domain sockets */
  if (proto == IMPULSE_TRANSPORT_SOCK_TCPL ||
      proto == IMPULSE_TRANSPORT_SOCK_UDPL ||
      offset == 0)
  {
    *name = impulse_unistring_create(address);
    impulse_transport_unlock(transport);
    return 0;
  }
  
  key = impulse_unistring_create_sub(address, 0, offset);
  impulse_unistring_prepend_utf(key, ":", 0);
  impulse_unistring_prepend(key, transport->protocol);
  
  if (!(record = impulse_hash_recover(rname_cache, key))) /* There is no cached entry */
  {
    impulse_unistring node = impulse_unistring_create_sub(address, 0, offset);
    impulse_transport_sock_dns_start(transport, sock_info[sock->protocol].family, node, &record, 1);
    impulse_hash_store(rname_cache, key, record);
    impulse_unistring_destroy(node);
    ret = IMPULSE_TRANSPORT_ADDRESS_RESOLVE;
  }
  else
  {
    if (record->inprogress) /* The cached entry is in progress */
      impulse_transport_sock_dns_end(record);
    
    if (!record->inprogress) /* The name lookup is complete */
    {
      if (record->data.entry.status < 0)
	ret = record->data.entry.status;
      else
      {
	*name = impulse_unistring_create_sub(address, offset, address->length);
	impulse_unistring_prepend(*name, record->data.entry.address);
      }
    }
    else
      ret = IMPULSE_TRANSPORT_ADDRESS_RESOLVE;
  }
  
  impulse_unistring_destroy(key);
  
  impulse_transport_unlock(transport);
  return ret;
}


impulse_int impulse_transport_sock_open(impulse_transport transport,
					impulse_unistring local,
					impulse_unistring remote)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_int ret;
  union sockaddr_all *laddr = NULL;
  union sockaddr_all *raddr = NULL;
  impulse_int laddr_len;
  impulse_int raddr_len;

  impulse_transport_lock(transport);
  if (sock->handle != -1)
    impulse_transport_sock_close(transport);

  if (local)
  {
    if ((ret = impulse_transport_sock_parse(sock->protocol, local, &laddr, &laddr_len, 1)) < 0)
    {
      impulse_transport_unlock(transport);
      return ret;
    }
    
    if (transport->local)
      impulse_unistring_assign(transport->local, local);
    else
      transport->local = impulse_unistring_create(local);
  }
  if (remote)
  {
    if ((ret = impulse_transport_sock_parse(sock->protocol, remote, &raddr, &raddr_len, 0)) < 0)
    {
      if (transport->local)
      {
	impulse_unistring_destroy(transport->local);
	transport->local = NULL;
      }
      impulse_transport_unlock(transport);
      return ret;
    }

    if (transport->remote)
      impulse_unistring_assign(transport->remote, remote);
    else
      transport->remote = impulse_unistring_create(remote);
  }
  
  if ((sock->handle = socket(sock_info[sock->protocol].family,
			     sock_info[sock->protocol].type, 0)) == -1)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create communications socket", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    /* Criticals never return */
  }

  ret = 1;
  if (setsockopt(sock->handle, SOL_SOCKET, SO_REUSEADDR, (void *)&ret, sizeof(ret)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot reuse socket address", 8);
    impulse_error_syscall(IMPULSE_ERROR_WARNING, desc);
    impulse_unistring_destroy(desc);
  }
  if ((ret = fcntl(sock->handle, F_GETFL)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot determine file handle flags", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }
  ret |= O_NONBLOCK;
  if (fcntl(sock->handle, F_SETFL, ret) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot enable non-blocking IO operations", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }
  
  if (local)
  {
    if ((ret = bind(sock->handle, &laddr->a, laddr_len)) < 0)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Call to bind failed", 8);
      impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
      impulse_unistring_destroy(desc);
    }
  }
  
  if (remote)
  {
    if ((ret = connect(sock->handle, &raddr->a, raddr_len)) < 0)
    {
      if (errno == EINPROGRESS)
      {
	ret = IMPULSE_TRANSPORT_CONNECTION_PENDING;
	impulse_event_register_fd(sock->handle,
				  IMPULSE_EVENT_FD_WRITE,
				  impulse_transport_sock_client,
				  NULL,
				  transport);
      }
      else
      {
	impulse_unistring desc = impulse_unistring_create_utf("Call to connect failed", 8);
	impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
	impulse_unistring_destroy(desc);
      }
    }

    if (!local)
    {
      union sockaddr_all addr;
      impulse_int addr_len = sizeof(addr);

      /*
       * Fix me: don't just return null.
       */
      if (getsockname(sock->handle, &addr.a, &addr_len) < 0)
      {
	printf("******** VERY BROKEN ******************\n");
	abort();
	impulse_transport_unlock(transport);
	return 0;
      }

      if (impulse_transport_sock_format(sock->protocol, &addr, addr_len, &transport->local) < 0)
      {
	printf("********** VERY BROKEN *****************\n");
	abort();
	impulse_transport_unlock(transport);
	return 0;
      }
    }

    if (!ret)
      impulse_event_register_fd(sock->handle,
				IMPULSE_EVENT_FD_READ | IMPULSE_EVENT_FD_EXCEPTION,
				impulse_transport_sock_client,
				NULL,
				transport);
  }
  else if (sock_info[sock->protocol].type == SOCK_STREAM)
  {
    if (!local)
    {
      impulse_unistring desc = impulse_unistring_create_utf("No local address specified for server", 8);
      impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, NULL);
      impulse_unistring_destroy(desc);
    }

    if (listen(sock->handle, 20) < 0) 
    {
      impulse_unistring desc = impulse_unistring_create_utf("Cannot listen on server socket", 8);
      impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
      impulse_unistring_destroy(desc);
    }
    impulse_event_register_fd(sock->handle,
			      IMPULSE_EVENT_FD_READ,
			      impulse_transport_sock_server,
			      NULL,
			      transport);
  }

  impulse_free(laddr);
  impulse_free(raddr);

  impulse_transport_unlock(transport);
  return ret;
}


static void impulse_transport_sock_incoming(impulse_transport transport,
					    impulse_int handle,
					    impulse_byte protocol)
{
  impulse_transport_sock sock = NULL;
  union sockaddr_all addr;
  impulse_int addr_len;
  impulse_int ret;
  
  sock = impulse_alloc(sizeof(struct _impulse_transport_sock));
  
  sock->handle = handle;
  sock->protocol = protocol;
  
  sock->handler = NULL;
  sock->edata = NULL;

  /* Fix me: don't return null, (free mem as well) */
  addr_len = sizeof(addr);
  if (getsockname(handle, &addr.a, &addr_len) < 0)
    ;

  if (impulse_transport_sock_format(sock->protocol, &addr, addr_len, &transport->local) < 0)
    ;
  
  addr_len = sizeof(addr);
  if (getpeername(handle, &addr.a, &addr_len) < 0)
    ;
  
  if (impulse_transport_sock_format(sock->protocol, &addr, addr_len, &transport->remote) < 0)
    ;

  if ((ret = fcntl(sock->handle, F_GETFL)) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot determine file handle flags", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }
  ret |= O_NONBLOCK;
  if (fcntl(sock->handle, F_SETFL, ret) < 0)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot enable non-blocking IO operations", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }

  impulse_event_register_fd(sock->handle,
			    IMPULSE_EVENT_FD_READ,
			    impulse_transport_sock_client,
			    NULL,
			    transport);
  
  transport->internal = sock;
}


impulse_int impulse_transport_sock_accept(impulse_transport transport,
					  impulse_transport *client)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_int handle = -1;

  impulse_transport_lock(transport);
  if (sock->handle == -1)
  {
    impulse_transport_unlock(transport);
    return IMPULSE_TRANSPORT_CONNECTION_CLOSED;
  }
  
  
  if ((handle = accept(sock->handle, NULL, NULL)) < 0)
  {
    if (errno != EAGAIN)
    {
      impulse_unistring desc = impulse_unistring_create_utf("Cannot accept incoming connection request", 8);
      impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
      impulse_unistring_destroy(desc);
    }
    impulse_transport_unlock(transport);
    return 0;
  }

  *client = impulse_alloc(sizeof(struct _impulse_transport));
  (*client)->impl = impulse_transport_load(transport->protocol);
  (*client)->protocol = impulse_unistring_create(transport->protocol);
  impulse_transport_sock_incoming(*client,
				  handle,
				  sock->protocol);
  impulse_transport_unlock(transport);
  return 1;
}


impulse_int impulse_transport_sock_close(impulse_transport transport)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  
  impulse_transport_lock(transport);
  if (sock->handle != -1)
  {
    impulse_event_deregister_fd(sock->handle);
    close(sock->handle);
    sock->handle = -1;
    
    if (transport->local)
      impulse_unistring_destroy(transport->local);
    if (transport->remote)
      impulse_unistring_destroy(transport->remote);
    
    transport->local = NULL;
    transport->remote = NULL;
  }
  
  impulse_transport_unlock(transport);
  return 0;
}


impulse_int impulse_transport_sock_active(impulse_transport transport)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_bool active;
  
  impulse_transport_lock(transport);
  
  active = sock->handle != -1;
  
  impulse_transport_unlock(transport);
  return active;
}


/* Data transfer functions */
impulse_int impulse_transport_sock_pending(impulse_transport transport)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  impulse_int pending;
  impulse_int ret;

  impulse_transport_lock(transport);

  if (sock->handle == -1)
  {
    impulse_transport_unlock(transport);
    return IMPULSE_TRANSPORT_CONNECTION_CLOSED;
  }
  
  if ((ret = ioctl(sock->handle, FIONREAD, &pending)) < 0)
  {
    /* Fix me:we might need to return CONNECTION_RESET sometimes */
    impulse_unistring desc = impulse_unistring_create_utf("Cannot determine pending data", 8);
    impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
    impulse_unistring_destroy(desc);
  }

  impulse_transport_unlock(transport);
  return pending;
}


impulse_int impulse_transport_sock_send(impulse_transport transport,
					impulse_byte *buffer,
					impulse_uint size,
					impulse_unistring address)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  union sockaddr_all *addr = NULL;
  impulse_int addr_len;
  impulse_int ret = 0;
  
  impulse_transport_lock(transport);
  if (sock->handle == -1)
  {
    impulse_transport_unlock(transport);
    return IMPULSE_TRANSPORT_CONNECTION_CLOSED;
  }
  
  if (address)
  {
    if ((ret = impulse_transport_sock_parse(sock->protocol,
					    address,
					    &addr,
					    &addr_len,
					    0)) < 0)
    {          
      impulse_transport_unlock(transport);
      return ret;
    }
  }
  
  restart:
  ret = sendto(sock->handle, buffer, size, 0, 
	       (address ? &addr->a : NULL), 
	       (address ? addr_len: 0));
  impulse_free(addr);

  if (ret < 0)
  {
    switch (errno)
    {
     case EAGAIN:
      return 0;
      
     case EINTR:
      goto restart;
      
     case EPIPE:
     case ECONNRESET:
      return IMPULSE_TRANSPORT_CONNECTION_RESET;
      
     case ENOTCONN:
      return IMPULSE_TRANSPORT_CONNECTION_PENDING;
      
     default:
      {
	impulse_unistring desc = impulse_unistring_create_utf("Cannot send data to socket", 8);
	impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
	impulse_unistring_destroy(desc);
      }
    }
  }
  impulse_transport_unlock(transport);
  return ret;
}


impulse_int impulse_transport_sock_receive(impulse_transport transport,
					   impulse_byte *buffer,
					   impulse_uint size,
					   impulse_unistring *address)
{
  impulse_transport_sock sock = (impulse_transport_sock)transport->internal;
  union sockaddr_all addr;
  impulse_int addr_len = sizeof(addr);
  impulse_int ret = 0;

  impulse_transport_lock(transport);
  if (sock->handle == -1)
  {    
    impulse_transport_unlock(transport);
    return IMPULSE_TRANSPORT_CONNECTION_CLOSED;
  }
  
  restart:
  if ((ret = recvfrom(sock->handle, (char *)buffer, size, 0, &addr.a, &addr_len)) == -1)
  {
    switch (errno)
    {
     case EAGAIN:
      impulse_transport_unlock(transport);
      return 0;
      
     case EINTR:
      goto restart;
      
     case ECONNRESET:
      impulse_transport_unlock(transport);
      return IMPULSE_TRANSPORT_CONNECTION_RESET;
      
     case ENOTCONN:
      impulse_transport_unlock(transport);
      return IMPULSE_TRANSPORT_CONNECTION_PENDING;
      
     default:
      {
	impulse_unistring desc = impulse_unistring_create_utf("Cannot send data to socket", 8);
	impulse_error_syscall(IMPULSE_ERROR_CRITICAL, desc);
	impulse_unistring_destroy(desc);
      }
    }
  }

  if (address)
  {
    if ((ret = impulse_transport_sock_format(sock->protocol,
					     &addr,
					     addr_len,
					     address)) < 0)
      impulse_transport_unlock(transport);
      return ret;
  }
  
  impulse_transport_unlock(transport);
  return ret;
}


#endif

/* End of file */
