/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * widget.c
 * --------
 *
 * Base widget class
 */

#include "../../../config.h"

#include "impulse/rows/main.h"
#include "impulse/rows/widget_int.h"
#include "impulse/util/memory.h"

/*                            ======--------======                           */

static void ROWS_widget_setup(ROWS_widget *widget,
			      int number);
static int  next_ID = 0;

/*                            ======--------======                           */

/* Base widget class functions */

/* Create the most boring widget ever */
ROWS_widget ROWS_widget_create(void)
{
  return *ROWS_widget_initialise(0);
}

/* Cast any widget type to our type */
ROWS_widget ROWS_widget_cast(void **widget)
{
  _p_ROWS_widget_info *info;

  info = widget[1];
  
  return *info->widget;
}

void ROWS_widget_destroy(void **widget)
{
  ROWS_widget          *wid;
  _p_ROWS_widget_info  *info;
  _p_ROWS_widget_data **data;
  int x;
  struct _p_ROWS_widget_notify_list *curr, *next;

  ROWS_widget_lock(widget);

  info = widget[1];
  wid  = info->widget;

  data = (_p_ROWS_widget_data **) ROWS_widget_private(wid, 0);

  curr = (*data)->destroy_notify;

  while (curr!=NULL)
    {
      next = curr->next;

      (curr->func)(*wid,
		   curr->destroy_data);

      impulse_free(curr);
      curr = next;
    }

  for (x=0; x<=(*data)->hierarchy_depth; x++)
    {
      info = wid[x][1];
      if (info->destroy != NULL)
	(info->destroy)(ROWS_widget_private(wid, x));

      impulse_free(wid[x][1]);
      impulse_free(wid[x]); 
      /* 
       * FIXME: or, CHECKME: when compiled with electric fence, the preceeding 
       *        line causes a segfault. I have no idea why, unless EF is too
       *        zealous about freeing things...
       *        (first iteration runs fine, second iteration segfaults on
       *         the for statement. Looks like a problem with data getting
       *         freed/corrupted)
       */
    }

#if USE_THREADS
  impulse_mutex_unlock((*data)->mutex);
  impulse_mutex_destroy((*data)->mutex);
#endif

  impulse_free(*data);
  impulse_free(wid);
}

/*                            ======--------======                           */

/* Derived widget support functions */

/* Gets a type ID for a given widget (only call once for any given widget!!) */
int ROWS_widget_next_type(void)
{
  int ID;

  impulse_mutex_global_lock;
  
  next_ID++;
  ID = next_ID;
  
  impulse_mutex_global_unlock;

  return ID;
}

/* Set our widget type ID */
void ROWS_widget_set_type(ROWS_widget *us,
			  int number,
			  int type_id)
{
  _p_ROWS_widget_info *info;

  ROWS_widget_lock(*us);
  
  info = us[number][1];
  info->type_id = type_id;

  ROWS_widget_unlock(*us);
}

/* The hierarchical number of the widget derived from this one */
int ROWS_widget_next_identifier(void)
{
  return 1; /* Widget has type number '1' */
}

/* Initialise widget structure */
ROWS_widget *ROWS_widget_initialise(int number)
{
  ROWS_widget         *widget;
  _p_ROWS_widget_info *info;
  int x;

  widget = impulse_alloc(sizeof(void **)*(number+1));
  
  for (x=0;x<=number;x++)
    {
      info = impulse_alloc(sizeof(_p_ROWS_widget_info));

      info->widget  = widget;
      info->type_id = -1;     /* Unidentified  */
      info->destroy = NULL;   /* No destructor */
      if (x==0)
	info->type_id = 0;
      info->destroy = NULL;
      
      widget[x]    = impulse_alloc(sizeof(void *)*2);
      widget[x][0] = NULL;
      widget[x][1] = info;
    }

  ROWS_widget_setup(widget, number);

  return widget;
}

/* Get our private word */
void **ROWS_widget_private(ROWS_widget *widget,
			   int number)
{
  return &(widget[number][0]);
}

/* Setup our widget */
static void ROWS_widget_setup(ROWS_widget *widget,
			      int number)
{
  _p_ROWS_widget_data **private;

  private  = (_p_ROWS_widget_data **) ROWS_widget_private(widget, 0);
  *private = impulse_alloc(sizeof(_p_ROWS_widget_data));

  (*private)->hierarchy_depth = number;
  (*private)->destroy_notify  = NULL;

#if USE_THREADS
  (*private)->mutex = impulse_mutex_create_recursive();
#endif
}

/* Cast any widget type to any other type that it is derived from */
void **ROWS_widget_cast_misc(void **widget, int number, int id)
{
  _p_ROWS_widget_info *info;
  ROWS_widget         *wid;

  if (number<0 || id == -1 || widget == NULL)
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, WIDGET_WRONGTYPE, id, widget);

  info = widget[1];
  wid  = info->widget;

  info = wid[number][1];

  if (info->type_id == -1)
    ROWS_error_raise(IMPULSE_ERROR_WARNING, WIDGET_INVALIDTYPE, id,
		     "Widget is untyped");
  else
  {
    if (info->type_id != id)
      ROWS_error_raise(IMPULSE_ERROR_CRITICAL, WIDGET_WRONGTYPE, id, widget);
  }

  return &(wid[number][0]);
}

void ROWS_widget_set_destructor (ROWS_widget *us,
				 int number,
				 ROWS_widget_destroy_func d)
{
  _p_ROWS_widget_info *info;

  ROWS_widget_lock(*us);
  
  info          = us[number][1];
  info->destroy = d;

  ROWS_widget_unlock(*us);
}

void ROWS_widget_destroy_notify (ROWS_widget us,
				 ROWS_widget_destroy_notify_func f,
				 void *d)
{
  _p_ROWS_widget_data **data;
  struct _p_ROWS_widget_notify_list *item;

  ROWS_widget_lock(us);
  
  data = (_p_ROWS_widget_data **) us;

  item               = impulse_alloc(sizeof(struct _p_ROWS_widget_notify_list));
  item->next         = (*data)->destroy_notify;
  item->func         = f;
  item->destroy_data = d;

  (*data)->destroy_notify = item;

  ROWS_widget_unlock(us);
}

void ROWS_widget_check_type(ROWS_widget widget,
			    int type,
			    char *expected_name)
{
  _p_ROWS_widget_info *info;

  if (widget==NULL)
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, WIDGET_INVALIDTYPE, type, expected_name);

  info = widget[1];

  if (info->type_id == -1)
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, WIDGET_INVALIDTYPE, type, "Widget is untyped");
  else
    if (info->type_id != type)
      ROWS_error_raise(IMPULSE_ERROR_CRITICAL, WIDGET_INVALIDTYPE, type, expected_name);
}

#if USE_THREADS
void ROWS_widget_lock(ROWS_widget widget)
{
  _p_ROWS_widget_info *info;
  _p_ROWS_widget_data **data;
  
  info = widget[1];
  data = (_p_ROWS_widget_data **) *info->widget;

  impulse_mutex_lock((*data)->mutex);
}

void ROWS_widget_unlock(ROWS_widget widget)
{
  _p_ROWS_widget_info *info;
  _p_ROWS_widget_data **data;
  
  info = widget[1];
  data = (_p_ROWS_widget_data **) *info->widget;

  impulse_mutex_unlock((*data)->mutex);
}
#endif
