/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Danie Berrange
 * are Copyright (C) 1999 Daniel Berrange
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * error.c
 * -------
 *
 * Impulse program error handling
 */

#include "../../../config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <syslog.h>
#include <signal.h>
#include <stdarg.h>

#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/thread.h"
#include "impulse/util/main.h"
#include "impulse/util/text.h"
#include "impulse/util/message.h"


static impulse_error_handler handlers[IMPULSE_ERROR_MAX+1] =
{
  impulse_error_bug, /* Emergencies */
  impulse_error_bug, /* Criticals */
  impulse_error_print, /* Warnings */
  NULL 
};

static impulse_uint messages[IMPULSE_ERROR_MAX+1] = 
{
  MSG_ERROR_EMERGENCY,
  MSG_ERROR_CRITICAL,
  MSG_ERROR_WARNING,
  MSG_ERROR_INFO
};

static impulse_uint priority[IMPULSE_ERROR_MAX+1] = 
{
  LOG_EMERG, LOG_CRIT, LOG_WARNING, LOG_INFO
};


/* 
 * Fix me: internationalise error messages 
 */

#if USE_THREADS
static impulse_bool impulse_dead = 0;
static impulse_mutex_static init_mutex = IMPULSE_MUTEX_CREATE_STATIC;
static impulse_mutex error_mutex = NULL;

static void impulse_error_init(void)
{
  impulse_mutex_lock(&init_mutex);
  if (!error_mutex)
    error_mutex = impulse_mutex_create_recursive();
  impulse_mutex_unlock(&init_mutex);
}
#endif



void impulse_error_raise(impulse_error_level level,
			 impulse_unistring format,
			 ...)
{
  va_list args;
  
  if (level < 0 || level > IMPULSE_ERROR_MAX)
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_ERROR_UNKNOWN));

#if USE_THREADS
  if (!error_mutex)
    impulse_error_init();
  impulse_mutex_lock(error_mutex);
  if (impulse_dead)
  {
    /* 
     * If another thread has already killed the process, then
     * we might as well get the hell out as well
     */
    impulse_mutex_unlock(error_mutex);
    _exit(1);
  }
#endif
  
  va_start(args, format);
  (*handlers[level])(level, format, args);
  va_end(args);
  if (level == IMPULSE_ERROR_EMERGENCY ||
      level == IMPULSE_ERROR_CRITICAL)
  {
    /* Do some application cleanup */
#if USE_THREADS
    impulse_dead = 1;
    impulse_mutex_unlock(error_mutex);
#endif
    exit(1);
  }
#if USE_THREADS
  impulse_mutex_unlock(error_mutex);
#endif
}


/*
 * Decide what happens according to error severity
 */
void impulse_error_handle(impulse_error_level level,
			  impulse_error_handler handler)
{
  if (level < 0 || level > IMPULSE_ERROR_MAX)
    ;
  //    impulse_error_raise(IMPULSE_ERROR_CRITICAL,
//			impulse_
    
#if USE_THREADS
  if (!error_mutex)
    impulse_error_init();
  impulse_mutex_lock(error_mutex);
#endif

  handlers[level] = handler;

#if USE_THREADS
  impulse_mutex_unlock(error_mutex);
#endif
}


void impulse_error_syscall(impulse_error_level level,
			   impulse_unistring format,
			   ...)
{
  va_list args;
  
  va_start(args, format);
  impulse_error_raise(level, format, args);
  va_end(args);
}


void impulse_error_print(impulse_error_level level,
			 impulse_unistring format,
			 ...)
{
  impulse_char *msg = impulse_unistring_to_8bit(format, impulse_app_charset());
  impulse_char *type = impulse_unistring_to_8bit(_(messages[level]), impulse_app_charset());
  
  fprintf(stderr, "***** %s\n", type);
  fprintf(stderr, "***** %s\n", msg);

  impulse_free(type);
  impulse_free(msg);
}


void impulse_error_log(impulse_error_level level,
		       impulse_unistring format,
		       ...)
{
  impulse_unistring app = impulse_app_name(IMPULSE_APP_NAME_PROGRAM);
  impulse_char *ident = impulse_unistring_to_8bit(app, impulse_app_charset());
  impulse_char *msg = impulse_unistring_to_8bit(format, impulse_app_charset());
  va_list vargs;
  
  va_start(vargs, format);
  
  openlog(ident, LOG_CONS | LOG_NDELAY | LOG_PERROR | LOG_PID, LOG_USER);
  syslog(priority[level], msg, vargs);
  closelog();
  
  impulse_free(msg);
  impulse_free(ident);
  va_end(vargs);
}


void impulse_error_bug(impulse_error_level level,
		       impulse_unistring format,
		       ...)
{
  impulse_uint child;
  impulse_uint pin[2];
  impulse_uint pout[2];
  impulse_unistring app = impulse_app_name(IMPULSE_APP_NAME_PROGRAM);
  impulse_unistring command = impulse_app_command();
  impulse_char in[20];
  impulse_char out[20];
  impulse_char *args[6];
  impulse_int status;
  
  va_list vargs;
  va_start(vargs, format);
  pipe(pin);
  pipe(pout);
  
  sprintf(in, "%d", pin[0]);
  sprintf(out, "%d", pout[1]);
  
  args[0] = "impulse_bug";
  args[1] = "--out";
  args[2] = out;
  args[3] = "--in";
  args[4] = in;
  args[5] = NULL;

  app = impulse_app_name(IMPULSE_APP_NAME_PROGRAM);
  
  child = fork();
  if (!child)
  {
    close(pin[1]);
    close(pout[0]);
    execvp(args[0], args);
    
    /* Only get here if the exec failed */
    impulse_error_print(level, format, vargs);
    _exit(0);
  }
  else if (child < 0)
  {
    impulse_error_print(level, format, vargs);
    return;
  }

  close(pin[0]);
  close(pout[1]);
  
  write(pin[1], &level, sizeof(level));
  
  write(pin[1], &app->length, sizeof(app->length));
  write(pin[1], app->data, sizeof(impulse_unichar)*(app->length+1));

  write(pin[1], &command->length, sizeof(command->length));
  write(pin[1], command->data, sizeof(impulse_unichar)*(command->length+1));

  write(pin[1], &format->length, sizeof(format->length));
  write(pin[1], format->data, sizeof(impulse_unichar)*(format->length+1));

  waitpid(child, &status, 0);

  close(pin[1]);
  close(pout[0]);
  va_end(vargs);
}


/* End of file */
