/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * list.c
 * ------
 *
 * Double headed lists
 */

#include "../../../config.h"

#include "impulse/util/memory.h"
#include "impulse/util/list.h"

/*                            ======--------======                           */

struct _impulse_list_bucket
{
  void *val;
  struct _impulse_list_bucket *next;
  struct _impulse_list_bucket *prev;
};


typedef struct _impulse_list_bucket *impulse_list_bucket;

struct _impulse_list
{
  impulse_list_bucket head;
  impulse_list_bucket iters;
  
  impulse_uint count;
  
  impulse_list_cleaner_func cleaner;
};


struct _impulse_list_iterator
{
  impulse_list list;
  impulse_list_bucket current;
};


typedef struct _impulse_list_iterator *impulse_list_iterator;

impulse_list_iterator impulse_list_iterator_create(impulse_list list);
void impulse_list_iterator_destroy(void *data);
impulse_bool impulse_list_iterator_eof(void *data);
void impulse_list_iterator_reset(void *data);
void *impulse_list_iterator_next(void *data);

static struct _impulse_iterator_funcs iterator_funcs = 
{
  impulse_list_iterator_destroy,
  impulse_list_iterator_eof,
  impulse_list_iterator_reset,
  impulse_list_iterator_next
};

/*                            ======--------======                           */

impulse_list impulse_list_create(void)
{
  impulse_list list = impulse_alloc(sizeof(struct _impulse_list));
  
  list->head = NULL;
  list->iters = NULL;
  list->count = 0;
  list->cleaner = NULL;

  return list;
}


void impulse_list_destroy(impulse_list list)
{
  impulse_list_bucket current, next;
  impulse_list_clear(list);
  
  /* Disconnect all iterators */
  current = list->iters;
  while (current)
  {
    impulse_list_iterator it = (impulse_list_iterator)current->val;
    it->list = NULL;
    next = current->next;
    impulse_free(current);
    current = next;
  }
  
  impulse_free(list);
}


void impulse_list_cleaner(impulse_list list,
			   impulse_list_cleaner_func func)
{
  list->cleaner = func;
}


impulse_iterator impulse_list_elements(impulse_list list)
{
  impulse_list_iterator it = impulse_list_iterator_create(list);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs, it);
  return iterator;
}


impulse_bool impulse_list_empty(impulse_list list)
{
  return list->count == 0;
}


impulse_uint impulse_list_count(impulse_list list)
{
  return list->count;
}


void impulse_list_append(impulse_list list,
			 void *data)
{
  impulse_list_bucket temp = impulse_alloc(sizeof(struct _impulse_list_bucket));
  
  temp->val = data;
  if (list->head)
  {
    temp->next = list->head;
    temp->prev = list->head->prev;
    list->head->prev->next = temp;
    list->head->prev = temp;
  }
  else
  {
    temp->next = temp;
    temp->prev = temp;
    list->head = temp;
  }
  list->count++;
}


void impulse_list_prepend(impulse_list list,
			  void *data)
{
  impulse_list_bucket temp = impulse_alloc(sizeof(struct _impulse_list_bucket));
  
  temp->val = data;
  if (list->head)
  {
    temp->next = list->head;
    temp->prev = list->head->prev;
    list->head->prev->next = temp;
    list->head->prev = temp;
    list->head = temp;
  }
  else
  {
    temp->next = temp;
    temp->prev = temp;
    list->head = temp;
  }
  list->count++;
}


void impulse_list_insert(impulse_list list,
			 impulse_uint index,
			 void *data)
{
  impulse_list_bucket temp = impulse_alloc(sizeof(struct _impulse_list_bucket));

  temp->val = data;
  if (!list->head)
  {
    temp->next = temp;
    temp->prev = temp;
    list->head = temp;
  }
  else
  {
    impulse_uint ind = 0;
    impulse_list_bucket current = list->head;
    while (ind < index)
    {
      current = current->next;
      ind++;
    }
    temp->next = current;
    temp->prev = current->prev;
    current->prev->next = temp;
    current->prev = temp;
    if (current == list->head)
      list->head = temp;
  }
  list->count++;
}


void *impulse_list_find(impulse_list list,
			impulse_uint index)
{
  impulse_list_bucket current = list->head;
  impulse_uint ind = 0;
  if ((index + 1) > list->count || index < 0)
    return NULL;
  
  while (ind < index)
  {
    current = current->next;
    ind++;
  }
  return current->val;
}


void *impulse_list_remove(impulse_list list,
			  impulse_uint index)
{
  impulse_list_bucket current = list->head;
  impulse_list_bucket iter = list->iters;
  impulse_uint ind = 0;
  void *data;
  
  if ((index + 1) > list->count || index < 0)
    return NULL;
  
  while (ind < index)
  {
    current = current->next;
    ind++;
  }
  
  /* Update any iterators pointing to the removed element */
  while (iter)
  {
    impulse_list_iterator it = (impulse_list_iterator)iter->val;
    if (it->current == current)
      it->current = (it->current == it->current->next) ? NULL : it->current->next;

    iter = iter->next;
  }
  
  current->next->prev = current->prev;
  current->prev->next = current->next;
  if (current == list->head)
    list->head = current->next;
  if (list->count == 1)
    list->head = NULL;
  
  data = current->val;
  impulse_free(current);
  list->count--;
  return data;
}


void impulse_list_clear(impulse_list list)
{
  impulse_list_bucket iter = list->iters;
  impulse_list_bucket current = list->head;
  
  /* Update all iterators currently active */
  while (iter)
  {
    impulse_list_iterator it = (impulse_list_iterator)iter->val;
    it->current = NULL;
    
    iter = iter->next;
  }
  
  if (list->head)
    list->head->prev->next = NULL;
  while (current)
  {
    impulse_list_bucket next = current->next;
    if (list->cleaner)
      (*list->cleaner)(current->val);
    impulse_free(current);
    current = next;
  }
  list->head = NULL;
  list->count = 0;
}


impulse_bool impulse_list_contains(impulse_list list,
				    void *data)
{
  impulse_list_bucket current = list->head;
  
  if (!list->head)
    return 0;
  
  while (current->val != data)
  {
    current = current->next;
    if (current == list->head)
      return 0;
  }
  return 1;
}


impulse_list_iterator impulse_list_iterator_create(impulse_list list)
{
  impulse_list_iterator it = impulse_alloc(sizeof(struct _impulse_list_iterator));
  impulse_list_bucket temp = impulse_alloc(sizeof(struct _impulse_list_bucket));
  
  it->list = list;
  temp->val = it;
  temp->next = it->list->iters;
  it->list->iters = temp;
  it->current = NULL;
  
  impulse_list_iterator_reset(it);
  
  return it;
}


void impulse_list_iterator_destroy(void *data)
{
  impulse_list_iterator it = (impulse_list_iterator)data;
  
  if (it->list)
  {
    impulse_list_bucket current = it->list->iters, prev = NULL;
    while (current)
    {
      if (current->val == it)
      {
	if (prev)
	  prev->next = current->next;
	else
	  it->list->iters = current->next;
	impulse_free(current);
	break;
      }
      prev = current;
      current = current->next;
    }
  }
  
  impulse_free(it);
}


impulse_bool impulse_list_iterator_eof(void *data)
{
  impulse_list_iterator it = (impulse_list_iterator)data;
  return it->current == NULL;
}


void impulse_list_iterator_reset(void *data)
{
  impulse_list_iterator it = (impulse_list_iterator)data;
  if (it->list)
    it->current = it->list->head;
  else
    it->current = NULL;
}


void *impulse_list_iterator_next(void *data)
{
  impulse_list_iterator it = (impulse_list_iterator)data;
  void *val = NULL;
  
  if (it->current)
  {
    val = it->current->val;
    it->current = it->current->next;
    if (it->current == it->list->head)
      it->current = NULL;
  }
    
  return val;
}


/* End of file */
