/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 */

/*
 * sparse.c
 * --------
 *
 * Sparse arrays
 */

#include "../../../config.h"

#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/util/sparse.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*                            ======--------======                           */
struct _impulse_sparse_bucket
{
  impulse_uint key;
  void *val;
  struct _impulse_sparse_bucket *next;
};


typedef struct _impulse_sparse_bucket *impulse_sparse_bucket;

struct _impulse_sparse
{
  impulse_sparse_bucket *cells;
  impulse_sparse_bucket iters;
  impulse_uint capacity;
  impulse_uint count;
  impulse_uint threshold;
  impulse_float load;
  
  impulse_sparse_cleaner_func cleaner;
};


struct _impulse_sparse_iterator
{
  impulse_sparse table;
  impulse_sparse_bucket current;
  impulse_uint cell;
  impulse_bool keys;
};


typedef struct _impulse_sparse_iterator *impulse_sparse_iterator;

impulse_sparse_iterator impulse_sparse_iterator_create(impulse_sparse table,
						   impulse_bool keys);
void impulse_sparse_iterator_destroy(void *data);
impulse_bool impulse_sparse_iterator_eof(void *data);
void impulse_sparse_iterator_reset(void *data);
void *impulse_sparse_iterator_next(void *data);

static struct _impulse_iterator_funcs iterator_funcs = 
{
  impulse_sparse_iterator_destroy,
  impulse_sparse_iterator_eof,
  impulse_sparse_iterator_reset,
  impulse_sparse_iterator_next
};


/*
 * Fix me: impulse_hash has an identical array of primes. Other
 * people might want primes as well. Think of a good way of making
 * this publically available.
 */
static impulse_uint capacity_primes[] = 
{
  11,
  19,
  37,
  73,
  109,
  163,
  251,
  367,
  557,
  823,
  1237,
  1861,
  2777,
  4177,
  6247,
  9371,
  14057,
  21089,
  31627,
  47431,
  71143,
  106721,
  160073,
  240101,
  360163,
  540217,
  810343,
  1215497,
  1823231,
  2734867,
  4102283,
  6153409,
  9230113,
  13845163,
};


/*                            ======--------======                           */

impulse_sparse impulse_sparse_create(void)
{
  impulse_sparse table = impulse_alloc(sizeof(struct _impulse_sparse));
  impulse_int i;
  
  table->capacity = capacity_primes[0];
  table->count = 0;
  table->load = 0.75f;
  table->threshold = (impulse_uint)((impulse_float)table->capacity * table->load);
  table->cells = impulse_alloc(sizeof(struct _impulse_sparse_bucket *) * table->capacity);
  table->iters = NULL;
  table->cleaner = NULL;
  
  for (i = 0 ; i < table->capacity ; i++)
    table->cells[i] = NULL;
  
  return table;
}


void impulse_sparse_destroy(impulse_sparse table)
{
  impulse_sparse_bucket current = table->iters, next;
  impulse_sparse_clear(table);
  
  /* Disconnect all iterators */
  while (current)
  {
    impulse_sparse_iterator it = (impulse_sparse_iterator)current->val;
    it->table = NULL;
    next = current->next;
    impulse_free(current);
    current = next;
  }
  
  impulse_free(table->cells);
  impulse_free(table);
}


void impulse_sparse_load(impulse_sparse table,
                       impulse_float factor)
{
  table->load = factor;
  impulse_sparse_pack(table);
}


void impulse_sparse_pack(impulse_sparse table)
{
  impulse_sparse_bucket *base;
  impulse_uint i, bigger = 0;

  for (i = 0; i < (sizeof(capacity_primes)/sizeof(impulse_uint)) && !bigger ; i++)
    if (capacity_primes[i] > table->capacity)
      bigger = capacity_primes[i];
  if (!bigger) /* This should never happen unles the table has about 13 million entries */
    bigger = (table->capacity * 2) + 1;
  
  base = impulse_alloc(sizeof(struct _impulse_sparse_bucket *) * bigger);
  for (i = 0 ; i < bigger ; i++)
    base[i] = NULL;
  for (i = 0 ; i < table->capacity ; i++)
  {
    impulse_sparse_bucket old = table->cells[i];
    while (old)
    {
      impulse_sparse_bucket next = old->next;
      impulse_uint index = (old->key & 0x7fffffff) % bigger;
      old->next = base[index];
      base[index] = old;
      old = next;
    }
  }
  impulse_free(table->cells);
  table->cells = base;
  table->capacity = bigger;
  table->threshold = (impulse_uint)((impulse_float)table->capacity * table->load);
}


void impulse_sparse_cleaner(impulse_sparse sparse,
			  impulse_sparse_cleaner_func func)
{
  sparse->cleaner = func;
}


impulse_iterator impulse_sparse_elements(impulse_sparse sparse)
{
  impulse_sparse_iterator it = impulse_sparse_iterator_create(sparse, 0);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs,
						      it);
  return iterator;
}


impulse_iterator impulse_sparse_keys(impulse_sparse sparse)
{
  impulse_sparse_iterator it = impulse_sparse_iterator_create(sparse, 1);
  impulse_iterator iterator = impulse_iterator_create(&iterator_funcs,
						      it);
  return iterator;
}


impulse_bool impulse_sparse_empty(impulse_sparse sparse)
{
  return sparse->count == 0;
}


impulse_uint impulse_sparse_count(impulse_sparse sparse)
{
  return sparse->count;
}


impulse_uint impulse_sparse_capacity(impulse_sparse sparse)
{
  return sparse->capacity;
}


void impulse_sparse_store(impulse_sparse table,
			  impulse_uint key,
			  void *val)
{
  impulse_uint index;
  impulse_sparse_bucket current, temp;

  if ((table->count + 1) > table->threshold)
    impulse_sparse_pack(table);
  
  index = (key & 0x7fffffff) % table->capacity;
  current = table->cells[index];
  while (current)
  {
    if (current->key == key)
    {
      if (table->cleaner)
	(*table->cleaner)(current->val);
      current->val = val;
      return;
    }
    current = current->next;
  }

  temp = impulse_alloc(sizeof(struct _impulse_sparse_bucket));
  temp->val = val;
  temp->key = key;
  temp->next = table->cells[index];
  table->cells[index] = temp;
  table->count++;
}


void *impulse_sparse_recover(impulse_sparse table,
			     impulse_uint key)
{
  impulse_uint index = (key & 0x7fffffff) % table->capacity;
  impulse_sparse_bucket current;

  current = table->cells[index];
  while (current)
  {
    if (current->key == key)
      return current->val;
    current = current->next;
  }

  return NULL;
}


void *impulse_sparse_remove(impulse_sparse table,
			    impulse_uint key)
{
  impulse_uint index = (key & 0x7fffffff) % table->capacity;
  impulse_sparse_bucket current, prev = NULL;
  impulse_sparse_bucket iter = table->iters;
  
  current = table->cells[index];
  while (current)
  {
    if (current->key == key)
    {
      void *data;
      if (prev)
        prev->next = current->next;
      else
        table->cells[index] = current->next;
      
      /* Disconnect all iterators */
      while (iter)
      {
	impulse_sparse_iterator it = (impulse_sparse_iterator)iter->val;
	if (it->current == current)
	  impulse_sparse_iterator_next(it);
	iter = iter->next;
      }
      
      data = current->val;
      impulse_free(current);
      table->count--;

      return data;
    }
    prev = current;
    current = current->next;
  }

  return NULL;
}


void impulse_sparse_clear(impulse_sparse table)
{
  impulse_sparse_bucket current = table->iters;
  impulse_int i;
  
  for (i = 0 ; i < table->capacity ; i++)
  {
    impulse_sparse_bucket current = table->cells[i];
    while (current)
    {
      impulse_sparse_bucket prev = current;
      current = current->next;
      if (table->cleaner)
	(*table->cleaner)(prev->val);
      impulse_free(prev);
    }
    table->cells[i] = NULL;
  }
  table->count = 0;
  
  while (current)
  {
    impulse_sparse_iterator it = (impulse_sparse_iterator)current->val;
    it->current = NULL;
    current = current->next;
  }
}


impulse_bool impulse_sparse_contains(impulse_sparse table,
				     void *data)
{
  impulse_int i;
  
  for (i = 0 ; i < table->capacity ; i++)
  {
    impulse_sparse_bucket current = table->cells[i];
    while (current)
    {
      if (current->val == data)
        return 1;
      current = current->next;
    }
  }
  
  return 0;
}


impulse_bool impulse_sparse_contains_key(impulse_sparse table,
					 impulse_uint key)
{
  impulse_uint index = (key & 0x7fffffff) % table->capacity;
  impulse_sparse_bucket current = table->cells[index];
  
  while (current)
  {
    if (current->key == key)
      return 1;
    current = current->next;
  }
  
  return 0;
}


impulse_sparse_iterator impulse_sparse_iterator_create(impulse_sparse table,
						       impulse_bool keys)
{
  impulse_sparse_iterator it = impulse_alloc(sizeof(struct _impulse_sparse_iterator));
  impulse_sparse_bucket temp = impulse_alloc(sizeof(struct _impulse_sparse_bucket));
  
  it->table = table;
  it->keys = keys;
  it->current = NULL;

  temp->val = it;
  temp->next = table->iters;
  table->iters = temp;
  
  impulse_sparse_iterator_reset(it);
  
  return it;
}


void impulse_sparse_iterator_destroy(void *data)
{
  impulse_sparse_iterator it = (impulse_sparse_iterator)data;
  
  if (it->table)
  {
    impulse_sparse_bucket current = it->table->iters, prev = NULL;
    while (current)
    {
      if (current->val == it)
      {
	if (prev)
	  prev->next = current->next;
	else
	  it->table->iters = current->next;
	impulse_free(current);
	break;
      }
      prev = current;
      current = current->next;
    }
  }

  impulse_free(it);
}


impulse_bool impulse_sparse_iterator_eof(void *data)
{ 
  impulse_sparse_iterator it = (impulse_sparse_iterator)data;
  
  return it->current == NULL;
}


void impulse_sparse_iterator_reset(void *data)
{
  impulse_sparse_iterator it = (impulse_sparse_iterator)data;
  
  if (it->table)
  {
    impulse_int i;
    for (i = 0 ; i < it->table->capacity && !it->current ; i++)
    {
      it->current = it->table->cells[i];
      it->cell = i;
    }
  }
  else
    it->current = NULL;
}


void *impulse_sparse_iterator_next(void *data)
{
  impulse_sparse_iterator it = (impulse_sparse_iterator)data;
  void *val;
  
  if (!it->current)
    return NULL;

  val = (it->keys) ? (void *)it->current->key : it->current->val;

  it->current = it->current->next;
  if (!it->current)
  {
    impulse_int i;
    for (i = it->cell + 1 ; i < it->table->capacity && !it->current ; i++)
    {
      it->current = it->table->cells[i];
      it->cell = i;
    }
  }

  return val;
}


/* End of file */
