/*                                                                |- Impulse -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the Impulse utility library
 * for Impulse. (Part of the ROWS distribution)
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange. Portions created by Daniel Berrange
 * are Copyright (C) 1999 Daniel Berrange.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * text.c
 * --------
 *
 * Internationalisation support
 */

#include "../../../config.h"

#include "impulse/util/text_int.h"
#include "impulse/util/memory.h"
#include "impulse/util/error.h"
#include "impulse/util/message.h"

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <string.h>


/*                            ======--------======                           */


impulse_text impulse_text_create(impulse_text base,
				 impulse_unistring locale)
{
  impulse_unistring filename;
  impulse_text text = NULL;
  impulse_char *locale_env = getenv("IMPULSE_TEXT_PATH");
  impulse_char *cfile;
  impulse_int fd;
  struct stat fs;
  impulse_int i;
  impulse_int length = 0;
  impulse_byte *map = NULL;
  impulse_uint nmessages = 0;

  
  /* 
   * $PREFIX/lib/impulse/text/<locale>/<package>.imt
   * $IMPULSE_TEXT_PATH/<locale>/<pacakge>.imt
   */
  
  if (locale_env)
  {
    filename = impulse_unistring_create_utf(locale_env, 8);
    impulse_unistring_append_utf(filename, "/", 8);
  }
  else
  {
    filename = impulse_unistring_create_utf(PREFIX_DIR, 8);
    impulse_unistring_append_utf(filename, "/lib/impulse/text/", 8);
  }
  impulse_unistring_append(filename, locale);
  impulse_unistring_append_utf(filename, "/", 8);
  impulse_unistring_append(filename, base->package);
  impulse_unistring_append_utf(filename, ".imo", 8);
  
  /* 
   * Damnn, the must be a better way of building the filename.
   * Unicode aware printf would be nice. Hmmm, I feel its time
   * to write an I/O module!      -- DPB 24/2/00
   */
  
  /*
   * Now load the data
   */

  cfile = impulse_unistring_to_utf(filename, 8, 0);
  impulse_unistring_destroy(filename);

  /*
   * This is a particularly sick bit of code, to progressively
   * strip off bits of the locale until we get a match.
   * 
   * Rewrite it sometime to make it look nice -- DPB 28/2/00
   */
  for (;;)
  {
    if ((fd = open(cfile, O_RDWR)) < 0)
    {
      impulse_char *sep, *dot, *under, *offset;
      
      sep = strrchr(cfile, '/');
      /* 
       * Blank off the end of the string, so that the search
       * for '.' & '_' does not pick up any in the filename
       * part of the string
       */
      if (sep)
	*sep = '\0';
      dot = strrchr(cfile, '.');
      under = strrchr(cfile, '_');
      offset = (dot && (dot > under)) ? dot : under;
      if (sep && offset)
      {
	*sep = '/'; /* Restore the dir separator */
	memmove(offset, sep, (strlen(cfile)-(sep-cfile))+1);
      }
      else
      {
	impulse_error_syscall(IMPULSE_ERROR_WARNING, _(MSG_TEXT_NOT_FOUND));
	impulse_free(cfile);
	return base;
      }
    }
    else
      break;
  }
  impulse_free(cfile);
  
  if (fstat(fd, &fs) < 0)
  {
    impulse_error_syscall(IMPULSE_ERROR_WARNING, _(MSG_TEXT_STAT_FAIL));
    close(fd);
    return base;
  }
  
  length = fs.st_size;
  if ((map = mmap(NULL, length, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0)) == MAP_FAILED)
  {
    impulse_error_syscall(IMPULSE_ERROR_WARNING, _(MSG_TEXT_MMAP_FAIL));
    close(fd);
    return base;
  }
  
  close(fd);

  /* First of all we must check the magic number */
  if (strcmp(map, "IMB")) 
  {
    impulse_error_raise(IMPULSE_ERROR_WARNING, _(MSG_TEXT_BAD_MAGIC));
    munmap(map, length);
    return base;
  }

  if ((nmessages = *(impulse_uint *)(map+4)) != base->nmessages) 
  {
    impulse_error_raise(IMPULSE_ERROR_WARNING, _(MSG_TEXT_WRONG_COUNT), nmessages);
    munmap(map, length);
    return base;
  }

  text = impulse_alloc(sizeof(struct _impulse_text));
  text->map = map;
  text->package = impulse_unistring_create(base->package);
  text->length = length;
  text->nmessages = nmessages;
  text->messages = (struct _impulse_unistring *)(text->map+8);
  
  /* Do pointer relocation on the data */
  for (i = 0 ; i < text->nmessages ; i++)
    (impulse_uint)text->messages[i].data += (impulse_uint)text->map;

  return text;
}


void impulse_text_destroy(impulse_text text)
{
  /* If there is no memory map, then we just have a pointer to the
   * static text object, so don't free anything */
  if (!text->map)
    return;

  munmap(text->map, text->length);
  impulse_unistring_destroy(text->package);
  impulse_free(text);
}


impulse_unistring impulse_text_message(impulse_text text,
				       impulse_uint id)
{
  if (id >= text->nmessages)
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, _(MSG_TEXT_TOKEN_RANGE), id, text->nmessages);
  
  return &text->messages[id];
}


/*                            ======--------======                           */


/* End of file */
