/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Andrew Hunter.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * thread_posix.c
 * --------------
 *
 * Impulse threading routines
 * POSIX Threads implementation
 */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>

/*                            ======--------======                           */
void impulse_thread_init(void)
{
}


void impulse_thread_cleanup(void)
{ 
}



impulse_thread impulse_thread_create(impulse_thread_main func,
				     void *argument)
{
  impulse_thread thread;

  THREAD_ERROR(pthread_create(&thread, NULL, func, argument));

  return thread;
}


impulse_thread impulse_thread_create_detach(impulse_thread_main func,
					    void *argument)
{
  impulse_thread thread;
  
  THREAD_ERROR(pthread_create(&thread, NULL, func, argument));
  THREAD_ERROR(pthread_detach(thread));
  
  return thread;
}


void impulse_thread_exit(void *status)
{
  pthread_exit(status);
}


void impulse_thread_join(impulse_thread t, void **status)
{
  THREAD_ERROR(pthread_join(t, status));
}


impulse_thread impulse_thread_self(void)
{
  return (impulse_thread)pthread_self();
}


impulse_bool impulse_thread_same(impulse_thread a,
				 impulse_thread b)
{
  return (pthread_equal(a, b) != 0);
}


void impulse_thread_nap(impulse_int secs,
			impulse_int usecs)
{
  /*
   * It may seem more logical to use sleep() here. However, sleep() is 
   * not thread-safe under most implementations of POSIX threads due
   * to the implementation of signals (the wake-up alarm signal can go 
   * to any thread, having very weird consequences). So, we use
   * select() instead, which is thread-safe usually.
   */
  struct timeval wakeup;

  wakeup.tv_sec  = secs;
  wakeup.tv_usec = usecs;
  
  select(0, NULL, NULL, NULL, &wakeup);
}


void impulse_thread_yield(void)
{
  sched_yield();
}


/*                            ======--------======                           */

impulse_mutex impulse_mutex_create(void)
{
# if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  impulse_mutex mutex = impulse_alloc(sizeof(pthread_mutex_t));
  THREAD_ERROR(pthread_mutex_init((pthread_mutex_t *)mutex, NULL));
# else
  impulse_mutex mutex = impulse_alloc(sizeof(impulse_mutex_impl));
  THREAD_ERROR(pthread_mutex_init(&mutex->mutex_native, NULL));

  mutex->is_recursive = 0;
  mutex->locked_by    = pthread_self();
  mutex->lock_level   = 0;
# endif
  
  return mutex;
}


impulse_mutex impulse_mutex_create_recursive(void)
{
#if HAVE_THREADS_POSIX_RECURSIVE
  pthread_mutexattr_t mutex_attr;
  impulse_mutex mutex = impulse_alloc(sizeof(pthread_mutex_t));

  THREAD_ERROR(pthread_mutexattr_init(&mutex_attr));
  THREAD_ERROR(pthread_mutexattr_settype(&mutex_attr,
					 PTHREAD_MUTEX_RECURSIVE));
  THREAD_ERROR(pthread_mutex_init((pthread_mutex_t*)mutex,
				  &mutex_attr));
  THREAD_ERROR(pthread_mutexattr_destroy(&mutex_attr));
#elif HAVE_THREADS_POSIX_LINUX_NP
  pthread_mutexattr_t mutex_attr;
  impulse_mutex mutex = impulse_alloc(sizeof(pthread_mutex_t));
  THREAD_ERROR(pthread_mutexattr_init(&mutex_attr));
  THREAD_ERROR(pthread_mutexattr_setkind_np(&mutex_attr,
					    PTHREAD_MUTEX_RECURSIVE_NP));
  THREAD_ERROR(pthread_mutex_init((pthread_mutex_t*)mutex,
				  &mutex_attr));
  THREAD_ERROR(pthread_mutexattr_destroy(&mutex_attr));
#else
  impulse_mutex mutex = impulse_alloc(sizeof(impulse_mutex_impl));
  THREAD_ERROR(pthread_mutex_init(&mutex->mutex_native, NULL));
  THREAD_ERROR(pthread_mutex_init(&mutex->mutex_recursive, NULL));

  mutex->is_recursive = 1;
  mutex->locked_by    = pthread_self();
  mutex->lock_level   = 0;
#endif
  return mutex;
}


void impulse_mutex_lock(impulse_mutex mutex)
{
#if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  /* We have recursive threads dealt with for us :-) */
  THREAD_ERROR(pthread_mutex_lock((pthread_mutex_t *)mutex));
#else
  if (mutex->is_recursive)
  {
    pthread_t self;
    
    self = pthread_self();
    
    /* We need exclusive access to the locking information */
    THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_recursive));
    
    /* If the lock level is zero, then we can lock the thread */
    if (mutex->lock_level == 0)
    {
      THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_native));
      
      mutex->lock_level++;
      mutex->locked_by = self;
    }
    else
    {
      if (mutex->locked_by == self)
      {
	mutex->lock_level++;
      }
      else
      {
	THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
	THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_native));
	THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_recursive));
	
	mutex->lock_level++;
	mutex->locked_by = self;
      }
    }
    THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
  }
  else
  {
    THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_native));
  }
#endif
}


impulse_bool impulse_mutex_trylock(impulse_mutex mutex)
{
#if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  /* We have recursive threads dealt with for us :-) */
  impulse_int ret;
  ret = pthread_mutex_trylock((pthread_mutex_t *)mutex);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
#else
  if (mutex->is_recursive)
  {
    pthread_t self = pthread_self();
    
    /* We need exclusive access to the locking information */
    THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_recursive));
    
    /* If the lock level is zero, then we can lock the thread */
    if (mutex->lock_level == 0)
    {
      THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_native));

      mutex->lock_level++;
      mutex->locked_by = self;
      THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
      return 1;
    }
    else
    {
      if (mutex->locked_by == self)
      {
	mutex->lock_level++;
	THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
	return 1;
      }
      else
      {
	THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
	return 0;
      }
    }
  }
  else
  {
    impulse_int ret;
    ret = pthread_mutex_lock(&mutex->mutex_native);
    if (ret == EBUSY)
      return 0;
    else if (ret < 0)
      THREAD_ERROR(ret);
    return 1;
  }
#endif
}

void impulse_mutex_unlock(impulse_mutex mutex)
{
#if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  /* We have recursive threads dealt with for us :-) */
  THREAD_ERROR(pthread_mutex_unlock((pthread_mutex_t*)mutex));
#else
  if (mutex->is_recursive)
  {
    pthread_t self;
    
    self = pthread_self();
    
    THREAD_ERROR(pthread_mutex_lock(&mutex->mutex_recursive));
    if (mutex->lock_level <= 0)
      impulse_error_raise(IMPULSE_ERROR_EMERGENCY, _(MSG_MUTEX_NOT_OWNER));
    if (mutex->locked_by != self)
      impulse_error_raise(IMPULSE_ERROR_EMERGENCY, _(MSG_MUTEX_ALREADY_LOCKED));
    
    if (mutex->lock_level > 1)
    {
      mutex->lock_level--;
      
      THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
    }
    else
    {
      mutex->lock_level = 0;
      THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_recursive));
      
      THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_native));
    }
    
    /* Note that mutex_recursive should be unlocked by now */
  }
  else
  {
    THREAD_ERROR(pthread_mutex_unlock(&mutex->mutex_native));
  }
#endif
}


void impulse_mutex_destroy(impulse_mutex mutex)
{
#if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  pthread_mutex_destroy((pthread_mutex_t*)mutex);
#else
  THREAD_ERROR(pthread_mutex_destroy(&mutex->mutex_native));
  if (mutex->is_recursive)
  {
    THREAD_ERROR(pthread_mutex_destroy(&mutex->mutex_recursive));
  }
#endif
  impulse_free(mutex);
}


static pthread_mutex_t global_lock = PTHREAD_MUTEX_INITIALIZER;

void impulse_mutex_f_global_lock(void)
{
  THREAD_ERROR(pthread_mutex_lock(&global_lock));
}

void impulse_mutex_f_global_unlock(void)
{
  THREAD_ERROR(pthread_mutex_unlock(&global_lock));
}

/*                            ======--------======                           */


impulse_cond impulse_cond_create(void)
{
  impulse_cond cond;

  cond = impulse_alloc(sizeof(pthread_cond_t));
  THREAD_ERROR(pthread_cond_init((pthread_cond_t*)cond, NULL));

  return cond;
}


void impulse_cond_signal(impulse_cond cond)
{
  THREAD_ERROR(pthread_cond_signal((pthread_cond_t*)cond));
}


void impulse_cond_broadcast(impulse_cond cond)
{
  THREAD_ERROR(pthread_cond_broadcast((pthread_cond_t*)cond));
}


void impulse_cond_wait(impulse_cond  cond,
		       impulse_mutex mutex)
{
#if HAVE_THREADS_POSIX_LINUX_NP || HAVE_THREADS_POSIX_RECURSIVE
  THREAD_ERROR(pthread_cond_wait((pthread_cond_t*)cond, (pthread_mutex_t*)mutex));
#else
  THREAD_ERROR(pthread_cond_wait((pthread_cond_t*)cond, &mutex->mutex_native));
#endif
}


void impulse_cond_destroy(impulse_cond cond)
{
  pthread_cond_destroy((pthread_cond_t*)cond);
  impulse_free(cond);
}


/*                            ======--------======                           */

/* Read/write mutexes */

impulse_rwlock impulse_rwlock_create(void)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  impulse_rwlock lock = impulse_alloc(sizeof(impulse_rwlock_impl));
  
  THREAD_ERROR(pthread_mutex_init(&lock->mutex, NULL));
  THREAD_ERROR(pthread_cond_init(&lock->notify, NULL));

  lock->readlocks = 0;
  lock->writelocks = 0;
#else
  impulse_rwlock lock = impulse_alloc(sizeof(pthread_rwlock_t));
  THREAD_ERROR(pthread_rwlock_init((pthread_rwlock_t*)lock, NULL));
#endif

  return lock;
}


void impulse_rwlock_read(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  pthread_mutex_lock(&lock->mutex);
  
  while (lock->writelocks > 0)
    THREAD_ERROR(pthread_cond_wait(&lock->notify, &lock->mutex));

  lock->readlocks++;

  THREAD_ERROR(pthread_mutex_unlock(&lock->mutex));
#else
  THREAD_ERROR(pthread_rwlock_rdlock((pthread_rwlock_t*)lock));
#endif
}


impulse_bool impulse_rwlock_tryread(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  pthread_mutex_lock(&lock->mutex);
  
  if (lock->writelocks > 0)
    return 0;
  
  lock->readlocks++;

  THREAD_ERROR(pthread_mutex_unlock(&lock->mutex));
  return 1;
#else
  impulse_int ret;
  ret = pthread_rwlock_tryrdlock((pthread_rwlock_t*)lock);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
#endif
}


void impulse_rwlock_write(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  THREAD_ERROR(pthread_mutex_lock(&lock->mutex));
  
  while (lock->writelocks > 0 && lock->readlocks > 0)
    THREAD_ERROR(pthread_cond_wait(&lock->notify, &lock->mutex));

  lock->writelocks++;

  THREAD_ERROR(pthread_mutex_unlock(&lock->mutex));
#else
  THREAD_ERROR(pthread_rwlock_wrlock((pthread_rwlock_t*)lock));
#endif
}


impulse_bool impulse_rwlock_trywrite(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  THREAD_ERROR(pthread_mutex_lock(&lock->mutex));
  
  if (lock->writelocks > 0 || lock->readlocks > 0)
    return 0;

  lock->writelocks++;

  THREAD_ERROR(pthread_mutex_unlock(&lock->mutex));
  return 1;
#else
  impulse_int ret;
  ret = pthread_rwlock_trywrlock((pthread_rwlock_t*)lock);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
#endif
}


void impulse_rwlock_unlock(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  THREAD_ERROR(pthread_mutex_lock(&lock->mutex));

  if (lock->readlocks > 0)
    lock->readlocks--;
  if (lock->writelocks > 0)
    lock->writelocks--;

  if (lock->readlocks <= 0 && lock->writelocks <= 0)
  {
    THREAD_ERROR(pthread_cond_broadcast(&lock->notify));
  }

  pthread_mutex_unlock(&lock->mutex);
#else
  pthread_rwlock_unlock((pthread_rwlock_t*)lock);
#endif
}


void impulse_rwlock_destroy(impulse_rwlock lock)
{
#if HAVE_THREADS_POSIX_RWLOCK==0
  pthread_mutex_destroy(&lock->mutex);
  pthread_cond_destroy(&lock->notify);
#else
  pthread_rwlock_destroy((pthread_rwlock_t*)lock);
#endif
  impulse_free(lock);
}

/* Semaphores */


pthread_mutex_t sema_mutex = PTHREAD_MUTEX_INITIALIZER;

static void _p_impulse_sema_create_static(impulse_sema sem)
{
  THREAD_ERROR(pthread_mutex_lock(&sema_mutex));
  if (!(*sem))
  {
    *sem = impulse_alloc(sizeof(sem_t));
    THREAD_ERROR(sem_init((sem_t*)*sem, 0, 0));
  }
  THREAD_ERROR(pthread_mutex_unlock(&sema_mutex));
}


impulse_sema impulse_sema_create(impulse_int value)
{
  impulse_sema sem = impulse_alloc(sizeof(sem_t *));
  *sem = impulse_alloc(sizeof(sem_t));
  
  THREAD_ERROR(sem_init((sem_t*)*sem, 0, value));
  
  return sem;
}


void impulse_sema_wait(impulse_sema sem)
{
  if (!(*sem))
    _p_impulse_sema_create_static(sem);
  
  THREAD_ERROR(sem_wait((sem_t*)*sem));
}


impulse_bool impulse_sema_trywait(impulse_sema sem)
{
  int ret;
  if (!(*sem))
    _p_impulse_sema_create_static(sem);
  
  ret = sem_trywait((sem_t*)*sem);
  if (ret == EBUSY)
    return 0;
  else if (ret < 0)
    THREAD_ERROR(ret);
  return 1;
}


void impulse_sema_post(impulse_sema sem)
{
  if (!(*sem))
    _p_impulse_sema_create_static(sem);
  
  THREAD_ERROR(sem_post((sem_t*)*sem));
}


void impulse_sema_destroy(impulse_sema sem)
{
  THREAD_ERROR(sem_destroy((sem_t*)*sem));
  impulse_free(*sem);
  impulse_free(sem);
}


/* End of file */
