#!/usr/bin/perl -w
#                                                                 |- Impulse -|
# The contents of this file are subject to the Mozilla
# Public License Version 1.0 (the "License"); you may
# not use this file except in compliance with the
# License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/ 
#
# Software distributed under the License is distributed
# on an "AS IS" basis, WITHOUT WARRANTY OF
# ANY KIND, either express or implied. See the
# License for the specific language governing rights
# and limitations under the License. 
#
# The Original Code is the ROWS widget library
# for Impulse.
#
# The Initial Developer of the Original Code is
# Daniel Berrange. Portions created by Daniel Berrange
# are Copyright (C) 2000 Daniel Berrange.
# All Rights Reserved
#
# Contributor(s):
# No-one (yet :-)
#


#
# Creates a C header file with message id defs from the master
# message database
#
# Syntax: msg2bin.pl [--translate locale] [--interface] <catalogue>


#
# This is a quick & dirty hack to allow us to get around a lovely
# little chicken & egg scenario:
# 
# The full message compiler uses libImpUtil, but in order to do
# so, the message databases have to be available!

use strict;

my $translate = "";
my $cfile = "";
my $hfile = "";
my $catalogue = "";
my $package = "";

my $i = 0;
for ($i = 0 ; $i <= $#ARGV ; $i++) {
  if ($ARGV[$i] eq "--translate") {
    if ($i < $#ARGV) {
      $translate = $ARGV[$i+1];
    } else { &help(); }
    $i++;
  } elsif ($ARGV[$i] eq "--cfile") {
    if ($i < $#ARGV) {
      $cfile = $ARGV[$i+1];
    } else { &help(); }
    $i++;
  } elsif ($ARGV[$i] eq "--hfile") {
    if ($i < $#ARGV) {
      $hfile = $ARGV[$i+1];
    } else { &help(); }
    $i++;
  } else {
    if ($catalogue eq "") {
      $catalogue = $ARGV[$i];
    } else {
      &help();
    }
  }
}

&help() if $catalogue eq "";

my @translations;
my @messages;

&process_catalogue($catalogue);

printf "Read %d translations, %d messages\n", $#translations+1, $#messages+1;

my $base = "";
my $offset = rindex $catalogue, "/";
if ($offset >= 0) {
  $base = substr $catalogue, 0, ($offset+1);
}

if ($translate ne "") {
  my $i;
  foreach ($i = 0 ; $i <= $#translations ; $i++) {
    my $locale = $translations[$i]->{"locale"};
    my $href = $translations[$i]->{"href"};

    &process_translation($locale, $href, $translate) if $translate eq "$locale.imo";
  }
}

if ($cfile ne "" || $hfile ne "") {
  &process_interface($cfile, $hfile);
}

print "Done\n\n";
exit;



sub process_catalogue {
  my $href = shift;

  open CATALOGUE, "<$href"  or die "Cannot open source file $href ($!)";

  my $lnum = 0;
  # Check for first line XML magic
  my $line = <CATALOGUE>;
  $lnum++;
  die "Unknown file type found" if ! ($line =~ /<\?xml\s+version=\"1.0\"\s+encoding=\"UTF-8\"\s+\?>/);
  
  # Check for doctype magic
  
  $line = <CATALOGUE>;
  $lnum++;
  die "Unknown file type found" if ! ($line =~ /<!DOCTYPE\s*impulse_message_catalogue/);
  
  do {
    $line = <CATALOGUE>;
    $lnum++;
    die "Package definition not found" if ! defined $line;
  } until $line =~ /<catalogue\s+name=\"(\w+)\"\s*>/;

  $package = $1;
  
  while (defined ($line = <CATALOGUE>)) {
    $lnum++;
    
    last if $line =~ /<\/catalogue>/;
    next if $line =~ /^\s*$/;
    
    if ($line =~ /<translation\s+.*\/>/) {
      my $locale = "";
      my $href = "";
      
      $locale = $1 if $line =~ /locale=\"([^\"]*)\"/;
      $href = $1 if $line =~ /href=\"([^\"]*)\"/;
      
      if ($locale eq "") {
        print "Warning: no locale specified for translation line $lnum $line";
        next;
      }
      
      if ($href eq "") {
        $href = $package . "_" . $locale;
      }
      
      push @translations, { "locale" => $locale, "href" => $href};
    
    } elsif ($line =~ /\s*<message\s+token=\"([^\"]+)\"([^>]*)>(.*)$/) {
      my $token = $1;
      my $content = $3;
      my $notranslate = ($2 =~ /notranslate/ ? 1 : 0);
      my $default = undef;
      my $done = 0;

      do {
        if (!defined ($default)) {
	  $default = "";
	} else {
	  $default .= "\n";
	}
        my $stuff;
        if ($content =~ /(.*)<\/message>/) {
          $stuff .= $1;
	  $done = 1;
	} else {
          $stuff .= $content;
        }
	if ($stuff =~ /^\s*\"(.*)\"\s*$/) {
          $default .= $1;
	} elsif ($stuff =~ /[^\s]/) {
	  $default .= $stuff;
	}
      } while (!$done && defined($content = <CATALOGUE>));

      die "Unexpected end of file" if (!$done);

      push @messages, { "token" => $token, "default" => $default, "notranslate" => $notranslate };
    
    } else {
      print "Warning: unknown entry line $lnum - $line";
    }
  }

  close CATALOGUE;
}



sub process_translation {
  my $locale = shift;
  my $href = shift;
  my $dest = shift;

  print "Translating $locale\n";
  open TRANSLATION, "<$base$href.imt"  or die "Cannot open source file $base$href.imt ($!)";

  my $lnum = 0;
  # Check for first line XML magic
  my $line = <TRANSLATION>;
  $lnum++;
  die "Unknown file type found" if ! ($line =~ /<\?xml version=\"1.0\" encoding=\"UTF-8\" \?>/);
  
  # Check for doctype magic
  
  $line = <TRANSLATION>;
  $lnum++;
  die "Unknown file type found" if ! ($line =~ /<!DOCTYPE impulse_message_translation/);
  
  do {
    $line = <TRANSLATION>;
    $lnum++;
    die "Package definition not found" if ! defined $line;
  } until $line =~ /<translation .*>/;

  my $translation = {};

  while (defined ($line = <TRANSLATION>)) {
    $lnum++;
    
    last if $line =~ /<\/translation>/;
    next if $line =~ /^\s*$/;
    
    if ($line =~ /<message .*\/>/) {
   } elsif ($line =~ /\s*<message\s+token=\"([^\"]+)\">(.*)$/) {
      my $token = $1;
      my $content = $2;
      my $default = undef;
      my $done = 0;
      
      do {
        if (!defined ($default)) {
	  $default = "";
	} else {
	  $default .= "\n";
	}
        my $stuff;
        if ($content =~ /(.*)<\/message>/) {
          $stuff .= $1;
	  $done = 1;
	} else {
          $stuff .= $content;
        }
	if ($stuff =~ /^\s*\"(.*)\"\s*$/) {
          $default .= $1;
	} elsif ($stuff =~ /[^\s]/) {
	  $default .= $stuff;
	}
      } while (!$done && defined($content = <TRANSLATION>));

      die "Unexpected end of file" if (!$done);

      $translation->{$token} = $default;
    } else {
      print "Warning: unknown entry line $lnum - $line";
    }
  }
  close TRANSLATION;
  
  open BINARY, ">$dest";

  my $magic = pack "a4", "IMB";
  print BINARY $magic;
  
  my $nstr = pack "I", $#messages+1;
  print BINARY $nstr;

  my $data = 4 + 4 + (($#messages+1) * 12);
  my $offset = 0;

  my $i;
  for ($i = 0 ; $i <= $#messages ; $i++) {
    my $text = $translation->{$messages[$i]->{"token"}} || $messages[$i]->{"default"};
    my $len = length $text;
    my $pos = $data + $offset;
    
    my $struct = pack "III", $len, $len, $pos;
    print BINARY $struct;
    
    $offset += 2 * ($len + 1);
  }
  for ($i = 0 ; $i <= $#messages ; $i++) {
    my $text = $translation->{$messages[$i]->{"token"}} || $messages[$i]->{"default"};
    my $len = length $text;
    
    my $struct = pack "S" . $len . "S", map{ord($_)} split (//, $text), 0;
    print BINARY $struct;
  }

  close BINARY;
}


sub process_interface {
  my $cfile = shift;
  my $hfile = shift;

  if ($hfile ne "") {
    print "Generating header file\n";
    open HEADER, ">$hfile" or die "Cannot create C header file $hfile ($!)";
   
    print HEADER "/*\n";
    print HEADER " * Automatically generated from $catalogue by msg2bin.pl\n";
    print HEADER " * Do NOT edit this file - all changes will be lost\n";
    print HEADER " * See $catalogue for copyright terms\n";
    print HEADER " */\n\n\n";

    print HEADER "#ifndef __$package" . "_text_h\n";
    print HEADER "#define __$package" . "_text_h\n\n";
    print HEADER "#include <impulse/util/text.h>\n\n";
    print HEADER "#ifdef __cplusplus\n";
    print HEADER "extern \"C\" {\n";
    print HEADER "#endif\n\n";

    print HEADER "extern impulse_text $package" . "_text;\n";
    print HEADER "extern impulse_text $package" . "_text_base;\n\n";

    print HEADER "#ifdef _\n";
    print HEADER "#undef _\n";
    print HEADER "#endif\n";
    print HEADER "#define _(x) impulse_text_message($package" . "_text, (x))\n\n";

    print HEADER "typedef enum _$package" . "_token {\n";
    my $i;
    for ($i = 0 ; $i <= $#messages ; $i++) {
      print HEADER "  " . $messages[$i]->{"token"} . ",\n";
    }
    print HEADER "} $package" . "_token;\n\n";
  
    print HEADER "#ifdef __cplusplus\n";
    print HEADER "}\n";
    print HEADER "#endif\n\n";
    print HEADER "#endif\n\n";
    print HEADER "/* End of file */\n";
    close HEADER;
  }
  
  if ($cfile ne "") {
    print "Generating source file\n";
    open SOURCE, ">$cfile" or die "Cannot create C source file $cfile ($!)";
    print SOURCE "/*\n";
    print SOURCE " * Automatically generated from $catalogue by msg2bin.pl\n";
    print SOURCE " * Do NOT edit this file - all changes will be lost\n";
    print SOURCE " * See $catalogue for copyright terms\n";
    print SOURCE " */\n\n\n";
    print SOURCE "#include \"impulse/util/unistring.h\"\n";
    print SOURCE "#include \"impulse/util/text_int.h\"\n\n";
    print SOURCE "#include \"message.h\"\n\n";
    print SOURCE "#include <stddef.h>\n";
  
    print SOURCE "static struct _impulse_unistring _$package" . "_name = {\n";
    printf SOURCE " %u,%u, (impulse_unichar[]){\n", length $package, length $package;
    my @letters = split //, $package;
    foreach (@letters) {
      print SOURCE "  (impulse_unichar)'$_',\n";
    }
    print SOURCE "  (impulse_unichar)'\\0'\n";
    print SOURCE "  }\n};\n\n";
  
    print SOURCE "static struct _impulse_unistring _$package" . "_messages[] = {\n";
    foreach (@messages) {
      &write_block($_->{"default"}, \*SOURCE);
    }
    print SOURCE "};\n";

    print SOURCE "static struct _impulse_text _$package" . "_text = {\n";
    print SOURCE "  &_$package" . "_name,\n";
    print SOURCE " 0, NULL,\n";
    print SOURCE "  "  . ($#messages+1) . ",\n";
    print SOURCE "  _$package" . "_messages\n";
    print SOURCE "};\n";
    print SOURCE "impulse_text $package" . "_text_base = &_$package" . "_text;\n";
    print SOURCE "impulse_text $package" . "_text = &_$package" . "_text;\n\n";

    print SOURCE "/* End of file */\n";
    close SOURCE;
  }
}

sub write_block {
  my $string = shift;
  my $file = shift;

  print $file "(struct _impulse_unistring){\n";
  printf $file "%u,%u, (impulse_unichar[]){\n", length $string, length $string;
  foreach (unpack "C*", $string) {
    print $file "  (impulse_unichar)$_,\n";
  }
  print $file "  (impulse_unichar)0\n";
  print $file "  }\n},\n";
}

sub help {
  print "syntax: msg2bin.pl [--translate locale] [--hfile name] [--cfile name] <catalogue>\n";
  exit;
}

# End of file
