#!/usr/bin/perl

#                                                                 |- Impulse -|
# The contents of this file are subject to the Mozilla
# Public License Version 1.0 (the "License"); you may
# not use this file except in compliance with the
# License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/ 
#
# Software distributed under the License is distributed
# on an "AS IS" basis, WITHOUT WARRANTY OF
# ANY KIND, either express or implied. See the
# License for the specific language governing rights
# and limitations under the License. 
#
# The Original Code is the ROWS widget library
# for Impulse.
#
# The Initial Developer of the Original Code is
# Andrew Hunter. Portions created by Andrew Hunter
# are Copyright (C) 1999 Andrew Hunter.
# All Rights Reserved
#
# Contributor(s):
# Daniel Berrange. Portions created by Daniel Berrange
# are Copyright (C) 1999 Daniel Berrange
# All Rights Reserved
#

#
# Surprisingly long Perl script for creating the impulse_unidata.c file from 
# a Unicode Character database file (www.unicode.org)
#

use IO::Handle;
use POSIX;

@rotator = ( '.', 'o', '0', 'O', '0', 'o', '.' );

#
# Optional characteristics
#

$include_comments  = 0;
$include_old_names = 0;
$include_new_names = 0;

$infile  = $ARGV[0];
$outfile = $ARGV[1];

print STDERR "unidata2c.pl v0.2 written by Andrew Hunter (updated by Daniel Berrange)\n";

# Read the Unidata file

open INFILE, "<$infile";

print STDERR "Reading unidata.txt (might be a bit slow, sorry). ";
flush STDERR;

$numchars = 0;
$rot      = 0;

my %category_map = (
 "Lu" => 1, "Ll" => 2, "Lt" => 3, "Lm" => 4, "Lo" => 5,
 "Nd" => 6, "Nl" => 7, "No" => 8, "Pc" => 9, "Pd" => 10,
 "Ps" => 11, "Pe" => 12, "Pi" => 13, "Pf" => 14, "Po" => 15,
 "Sm" => 16, "Sc" => 17, "Sk" => 18, "So" => 19, "Mn" => 20,
 "Mc" => 21, "Me" => 22, "Zs" => 23, "Zl" => 24, "Zp" => 25,
 "Cc" => 26, "Cf" => 27, "Cs" => 28, "Co" => 29, "Cn" => 30
);

my %direction_map = (
 "L" => 1, "LRE" => 1, ,"LRO" => 1, "R" => 2, "AL" => 2,
 "RLE" => 2, "RLO" => 2, "EN" => 3, "ES" => 4, "ET" => 5,
 "AN" => 6, "CS" => 7, "B" => 8, "S" => 9, "WS" => "10",
 "ON" => 11
);

my %formatting_map = (
 "font" => 1, "noBreak" => 2, "initial" => 3, "medial" => 4, 
 "final" => 5, "isolated" => 6, "circle" => 7, "super" => 8,
 "sub" => 9, "vertical" => 10, "wide" => 11, "narrow" => 12,
 "small" => 13, "square" => 14, "fraction" => 15, "compat" => 16
);
 
while (<INFILE>)
  {
    if (/^([0-9A-F]+);(.*);(.*);([0-9]*);(L|R|EN|ES|ET|AN|CS|B|S|WS|ON);(.*);([0-9]*);([0-9]*);(.*);(Y|N);(.*);(.*);([0-9A-F]*);([0-9A-F]*);([0-9A-F]*)$/)
      {
	$numchars++;
	if (($numchars%512) == 0)
	  {
	    print STDERR "\b.$rotator[$rot]";
	    flush STDERR;
	  }
	if (($numchars%64) == 0)
	  {
	    $rot = ($rot+1)%($#rotator);
	    print STDERR "\b$rotator[$rot]";
	    flush STDERR;
	  }

	$field[0]  = $1;
	$field[1]  = $2;
	$field[2]  = $3;
	$field[3]  = $4;
	$field[4]  = $5;
	$field[5]  = $6;
	$field[6]  = $7;
	$field[7]  = $8;
	$field[8]  = $9;
	$field[9]  = $10;
	$field[10] = $11;
	$field[11] = $12;
	$field[12] = $13;
	$field[13] = $14;
	$field[14] = $15;

	$unicode  = hex($field[0]);
	$name     = $field[1];

	$category = 0;
	$category_i = 0;

        $category = $category_map{$field[2]} || 0;
	$combining = int($field[3]);

	$direction = $direction_map{$field[4]} || 0;

	if ($field[5] =~ /^(<(.*)> )*(([0-9A-F]+ *)*)$/)
	  {
	    $formatting = $formatting_map{$2} || 0;

	    $equiv = $3;
	    $num   = 0;
	    $items = [ ];
	    while ($equiv =~ /^([0-9A-F]+) *(.*)$/)
	      {
		$$items[$num] = hex($1);
		$num++;

		$equiv = "$2";
	      }

	    $decomposition = { length        => $num,
			       decomposition => $items,
			       formatting    => $formatting };
	  }
	else
	  {
	    die "Syntax error in Unicode character database at line $.\n";
	  }

	$decimalval = int($field[6]);
	$digitval   = int($field[7]);
	if ($field[8] =~ /^([0-9]+)\/([0-9]+)$/)
	  {
	    $numericval = $1 / $2;
	  }
	else
	  {
	    $numericval = int($field[8]);
	  }

	$ismirrored = 0;
	$ismirrored = 1 if ($field[9] eq "Y");

	$oldname = $field[10];
	$comment = $field[11];
	$uppercase = hex($field[12]);
	$lowercase = hex($field[13]);
	$titlecase = hex($field[14]);

	$unidata[$unicode] = { code            => $unicode,
			       name            => $name,
			       category        => $category,
			       combining       => $combining,
			       direction       => $direction,
			       decomposition   => $decomposition,
			       decimal_value   => $decimalval,
			       digit_value     => $digitval,
			       numeric_value   => $numericval,
			       is_mirrored     => $ismirrored,
			       old_name        => $oldname,
			       comment         => $comment,
			       uppercase       => $uppercase,
			       lowercase       => $lowercase,
			       titlecase       => $titlecase
			     };
	$occupied[$unicode/256] = 1;
      }
    else
      {
        die "Syntax error in Unicode character database at line $.\n";
      }
  }

close INFILE;

print STDERR "\b \b";
flush STDERR;

$totalspaces = 0;
for ($section=0;$section<256;$section++)
  {
    if ($occupied[$section] == 1)
      {
	$totalspaces += 256;
      }
  }

print STDERR "\nUnidata: processed $numchars definitions \($#unidata allocations\)\n";
#$estimate = ceil((32*$totalspaces)/1024);
$estimate = ceil((20 * $totalspaces) / 1024);
print STDERR "C: $totalspaces characters. Estimating $estimate\k object file\n";

# Generate composition data

print STDERR "Generating binary compositions... (   %)";
flush STDERR;

$binary_compo = 0;
$nary_compo   = 0;

for ($section=0; $section<256; $section++)
  {
    if ($occupied[$section])
      {
	$start = $section*256;
	$end   = $start + 256;

	for ($char=$start; $char<$end; $char++)
	  {
	    $info = $unidata[$char];
	    if (defined($$info{'decomposition'}))
	      {
		$decomp = $$info{'decomposition'};

		if ($$decomp{'length'} == 2 && $$decomp{'formatting'} == 0)
		  {
		    $set = $$decomp{'decomposition'};
		    $one = $$set[0];
		    $two = $$set[1];
		    if ($#$set != 1)
		      {
			print STDERR "Whoops $#$set\n";
		      }
		    $hash = "map-$one\-$two";

		    my $defn;

		    if (defined($comp_table{$hash}))
		      {
			$defn = $comp_table{$hash};
		      }
		    else
		      {
			$defn = { };
		      }

		    $$defn{'unichar'} = "$char";

		    # Store in a perl table
		    $comp_table{$hash} = $defn;
		    
		    # Store in the final representation
		    addcomposition(\@ctable, 16, ($one<<16)|$two, $defn);

		    $binary_compo++;
		  }
		else
		  {
		    if ($$decomp{'length'} > 2 && $$decomp{'formatting'} == 0)
		      {
			$set = $$decomp{'decomposition'};
			$one = $$set[0];
			$two = $$set[1];
			$hash = "map-$one\-$two";

			$nary[$nary_compo] = $char;
			$nary_compo++;
			if (!defined($comp_table{$hash}))
			  {
			    my $defn = { };
			    $comp_table{$hash} = $defn;
			  }
		      }
		  }
	      }
	  }
      }

    $percentage = floor(($section*100)/255);
    print STDERR "\b\b\b\b\b";
    print STDERR " " if ($percentage<10);
    print STDERR " " if ($percentage<100);
    print STDERR "$percentage\%)";
    flush STDERR;
  }

$percentage = floor(($nary_compo*100)/($binary_compo+$nary_compo));
print STDERR "\n$binary_compo 'pure' binary compositions processed. $nary_compo n-ary compositions left ($percentage\%)\n";

print STDERR "Genrating n-ary compositions...";
flush STDERR;

$tri     = 0;

for ($c = 0; $c<$nary_compo; $c++)
  {
    $char = $nary[$c];

    $info = $unidata[$char];
    $decomp = $$info{'decomposition'};
    
    $set   = $$decomp{'decomposition'};
    $one   = $$set[0];
    $two   = $$set[1];

    $hash = "map-$one\-$two";

    my $defn = { };

    if (defined($comp_table{$hash}))
      {
	$defn = $comp_table{$hash};
      }
    else
      {
	$defn = { };
      }
    $multi = $$defn{'multi'};
    $multito = $$defn{'multi_compo'};
    $num = $#$multi+1;
    $$multi[$num] = $set;
    $$multito[$num] = $char;

    $$defn{'multi'} = $multi;
    $$defn{'multi_compo'} = $multito;
 
    # Store in a perl table
    $comp_table{$hash} = $defn;
    
    # Store in the final representation
    addcomposition(\@ctable, 16, ($one<<16)|$two, $defn);

#    $stage_one   = $ctable[$one/256];
#    $stage_two   = $$stage_one[$one%256];
#    $stage_three = $$stage_two[$two/256];
#    $$stage_three[$two%256] = $defn;
#    $$stage_two[$two/256]   = $stage_three;
#    $$stage_one[$one%256]   = $stage_two;
#    $ctable[$one/256]       = $stage_one;
  }

# Create the C file

print STDERR "\nCreating impulse_unidata.c...\n";
print STDERR "Writing Character Databse... (   %)";
flush STDERR;

open CFILE, ">$outfile";

print CFILE "/*\n * File automatically generated by unidata2c.pl\n * Data derived from the Unicode Character database\n * www.unicode.org\n */\n\n";
print CFILE "/*\n * unidata2c.pl was written by Andrew Hunter\n */\n\n";

print CFILE "#include <stdio.h>\n";
print CFILE "#include \"impulse/util/unidata.h\"\n\n";
print CFILE "/*                            ======--------======                           */\n\n";
# print CFILE "impulse_unidata_item **impulse_unidata = (impulse_unidata_item **)\n  {\n";

$charon = 0;

printf CFILE "static impulse_unidata_item range_empty[] =\n";
printf CFILE "    {\n";
for ($char = 0; $char < 256; $char++)
  {
    printf CFILE "      {\n";
    printf CFILE "        0,0,0,0,0,0, NULL, 0,0,0, 0, 0.0\n";
    if ($char<255)
      {
	printf CFILE "      },\n";
      }
    else
      {
	printf CFILE "      }\n";
      }
  }
printf CFILE "    };\n\n";

for ($section = 0; $section < 256; $section++)
  {
    $start = $section * 256;
    $end   = $start   + 256;

    if ($occupied[$section] == 1)
      {
	printf CFILE "    /* Character range U+%04x to U+%04x */\n", $start, ($end-1);
	printf CFILE "static impulse_unidata_item range_U%04x[] =\n", $start;

	print CFILE "    {\n";

	for ($char = $start; $char<$end; $char++)
	  {
	    $info = $unidata[$char];
	    $decomp = $$info{'decomposition'};
	    print CFILE "      {\n";

	    if ($$info{'code'} == $char)
	      {
		$charon++;
	      }

	    $infostr = "";
	    $part1 = "";
	    $part2 = "";
	    $part3 = "";
	    if ($include_new_names != 0)
	      {
		$part1 = $$info{'name'};
	      }
	    if ($include_old_names != 0)
	      {
		$part2 = $$info{'old_name'};
	      }
	    if ($include_comments != 0)
	      {
		$part3 = $$info{'comment'};
	      }
	    $infostr = "$part1;$part2;$part3";
	    printf CFILE " %u,",      $$info{'category'};
	    printf CFILE " %u,",      $$info{'is_mirrored'};
	    printf CFILE " %u,",      $$info{'direction'};
	    if (defined $decomp) {
 	      printf CFILE " %u,",      $$decomp{'formatting'};
	    } else {
	      printf CFILE " 0,";
	    }
	    printf CFILE " %u,",      $$info{'combining'};
	    
	    if (defined $decomp && $$decomp{'length'} > 0) {
	      printf CFILE "          %u,", $$decomp{'length'};
	      printf CFILE "          (impulse_unichar[])\n          {\n            ";
	      $set = $$decomp{'decomposition'};
	      for ($sub=0; $sub<=$#$set; $sub++)
	      {
  	        printf CFILE "0x%04x", $$set[$sub];
	        if ($sub<$#$set)
		{
		  printf CFILE ", ";
		}
		else
		{
		  printf CFILE "\n";
		} 
	      }
	      printf CFILE "          },\n";
	    } else {
  	      printf CFILE " 0, NULL,\n";
	    }

	    printf CFILE "        0x%04x,",  $$info{'uppercase'};
	    printf CFILE " 0x%04x,",         $$info{'lowercase'};
	    printf CFILE " 0x%04x,\n",       $$info{'titlecase'};

	    printf CFILE " %i,",             $$info{'decimal_value'};
	    printf CFILE "        %1f\n",   $$info{'numeric_value'};

	    if ($char < ($end-1))
	      {
		print CFILE "      },\n";
	      }
	    else
	      {
		print CFILE "      }\n";
	      }
	  }

	print CFILE "    };\n";
      }

    $percentage = ceil(($charon*100) / $numchars);
    if ($percentage < 10)
      {
	$percentage = "  $percentage";
      }
    elsif ($percentage < 100)
      {
	$percentage = " $percentage";
      }
    print STDERR "\b\b\b\b\b$percentage%)";
    flush STDERR;
  }

print CFILE "\nimpulse_unidata_item *impulse_unidata[] =\n  {\n";

for ($section = 0; $section < 256; $section++)
  {
    $start = $section * 256;
    $end   = $start   + 256;

    if ($occupied[$section] == 1)
      {
	printf CFILE "    range_U%04x", $start;
      }
    else
      {
	printf CFILE "    range_empty";
      }

    if ($section < 255)
      {
	printf CFILE ",\n";
      }
    else
      {
	printf CFILE "\n";
      }
  }

printf CFILE "  };\n\n";

print STDERR "\n";

print STDERR "Writing composition table...\n";
writecompositions(\@ctable, 16, "impulse_unidata_comp_table"); 

# That's all folks

close CFILE;
print STDERR "That seems to be it\n";

sub addcomposition
  {
    my ($array, $level, $code, $entry) = @_;

    if ($level <= 1)
      {
	$$array[$code] = $entry;
      }
    else
      {
	my ($subcode)  = ($code%4);
	if (!defined($$array[$subcode]))
	  {
	    $$array[$subcode] = [ ];
	  }
	my ($newarray, $newcode) = ($$array[$subcode], $code/4);

	addcomposition($newarray, $level-1, $newcode, $entry);
      }
  }

sub writecompositions
  {
    my ($items, $level, $prefix) = @_;
    my ($item);

    if ($level <= 1)
      {
	print CFILE "static impulse_unidata_composition $prefix\[4] =\n{\n";
	for ($item=0; $item<4; $item++)
	  {
	    print CFILE "  {\n";
	    if (defined($$items[$item]))
	      {
		my ($desc) = $$items[$item];
		my ($flags) = 0;

		if (defined($$desc{'unichar'}))
		  {
		    $flags |= 1;
		  }
		if (defined($$desc{'multi'}))
		  {
		    $flags |= 2;
		  }

		print  CFILE "    $flags,\n";
		printf CFILE "    0x%04x,\n", $$desc{'unichar'};

		if (defined($$desc{'multi'}))
		  {
		    $multi = $$desc{'multi'};
		    $multito = $$desc{'multi_compo'};
		    printf CFILE "    %i,\n", $#$multi + 1;
		    printf CFILE "    (impulse_unichar[%i])\n    {\n      ", $#$multi + 1;
		    for ($x=0; $x<=$#$multi; $x++)
		      {
			printf CFILE "0x%04x", $$multito[$x];
			if ($x<$#$multi)
			  {
			    print CFILE ", ";
			  }
			else
			  {
			    print CFILE "\n";
			  }
		      }
		    print  CFILE "    },\n";
		    printf  CFILE "    (struct _multi_data[%i])\n    {\n", $#$multi+1;
		    for ($x=0; $x<=$#$multi; $x++)
		      {
			print CFILE "      {\n";
			$set = $$multi[$x];
			printf CFILE "        %i,\n", $#$set+1;
			printf CFILE "        (impulse_unichar[%i])\n        {\n          ", $#$set+1;
			for ($y=0; $y<=$#$set; $y++)
			  {
			    printf CFILE "0x%04x", $$set[$y];
			    if ($y == $#$set)
			      {
				print CFILE "\n";
			      }
			    else
			      {
				print CFILE ", ";
			      }
			  }
			print  CFILE "        }\n";

			if ($x == $#$multi)
			  {
			    print CFILE "      }\n";
			  }
			else
			  {
			    print CFILE "      },\n";
			  }
		      }
		    print  CFILE "    }\n";
		  }
		else
		  {
		    print  CFILE "    0,\n    NULL,\n    NULL\n";
		  }
	      }
	    else
	      {
		print CFILE "    0, 0, -1, NULL, NULL\n";
	      }

	    if ($item<3)
	      {
		print CFILE "  },\n";
	      }
	    else
	      {
		print CFILE "  }\n";
	      }
	  }

	print CFILE "};\n\n";
      }
    else
      {
	for ($item=0; $item<4; $item++)
	  {
	    if (defined($$items[$item]))
	      {
		my ($newitem)   = $$items[$item];
		my ($newprefix);
		if ($level == 16)
		  {
		    $newprefix = sprintf("c_%01x", $item);
		  }
		else
		  {
		    $newprefix = sprintf("%s%01x", $prefix, $item);
		  }
		writecompositions($newitem, $level-1, $newprefix);
	      }
	  }

	if ($level < 16)
	  {
	    print CFILE "static ";
	  }
	print CFILE "impulse_unidata_composition ";
	for ($x=1; $x<$level; $x++)
	  {
	    print CFILE "*";
	  }
	print CFILE "$prefix\[4] = \n{ ";
	if ($level == 16)
	  {
	    $prefix = "c_";
	  }
	for ($item=0; $item<4; $item++)
	  {
	    if (defined($$items[$item]))
	      {
		my ($newprefix) = sprintf("%s%01x", $prefix, $item);
		print CFILE "$newprefix";
	      }
	    else
	      {
		print CFILE "NULL";
	      }

	    if ($item<3)
	      {
		print CFILE ", ";
	      }
	    else
	      {
		print CFILE " ";
	      }
	  }
	print CFILE "};\n\n";
      }
  }
